# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This files contains IO operations on Feature related data.
Features are processed data linked to one or multiple record data.
For example Keypoints are features related to 1 RecordCamera (image).
And Matches are features related to a pair of RecordCamera (2 images).
"""

import numpy as np
import os.path as path
from typing import Tuple, Any, Dict, Type, Optional, Union, Iterable

import kapture
from kapture.io.records import RECORD_DATA_DIRNAME
from kapture.utils.paths import path_secure, populate_files_in_dirpath
from kapture.io.binary import array_from_file, array_to_file


# Feature files related functions ######################################################################################
FEATURES_DATA_DIRNAMES = {
    kapture.Keypoints: path.join('reconstruction', 'keypoints'),
    kapture.Descriptors: path.join('reconstruction', 'descriptors'),
    kapture.GlobalFeatures: path.join('reconstruction', 'global_features'),
    kapture.Matches: path.join('reconstruction', 'matches'),
    kapture.RecordsBase: RECORD_DATA_DIRNAME,  # prefer kapture.RecordsCamera, Lidar, ...
}

FEATURE_FILE_EXTENSION = {
    kapture.Keypoints: '.kpt',
    kapture.Descriptors: '.desc',
    kapture.GlobalFeatures: '.gfeat',
    kapture.Matches: '.matches',
}

FEATURE_PAIR_PATH_SEPARATOR = {
    kapture.Matches: '.overlapping'
}


# get file path for binary files in kapture ############################################################################
def get_features_fullpath(
        data_type: Any,
        kapture_dirpath: str = '',
        image_filename: Optional[str] = None) -> str:
    """
    Returns full path to subdirectory containing the binary files of the given type.
            Optionally, can give a the image file name, and add the feature file name (with proper extension).

    :param data_type:
    :param kapture_dirpath: input path to kapture directory.
    :param image_filename: optional input image filename (id).
    :return: Feature full path
    """
    subdir = FEATURES_DATA_DIRNAMES[data_type]
    feature_filename = image_filename + FEATURE_FILE_EXTENSION[data_type] if image_filename else ''
    return path_secure(path.join(kapture_dirpath, subdir, feature_filename))


def features_to_filepaths(
        kapture_data: Union[kapture.RecordsCamera,
                            kapture.Keypoints,
                            kapture.Descriptors,
                            kapture.GlobalFeatures],
        kapture_dirpath: str = '') -> Dict[str, str]:
    """
    Returns a dict mapping image_id to path to feature file.
        Eg.{image_id: full/path/to/feature.jpg.kpt}

    :param kapture_data: input kapture data.
    :param kapture_dirpath: input path to kapture directory.
    :return: image id to image file path dictionary
    """
    data_type = type(kapture_data)
    return {
        image_filename: get_features_fullpath(data_type, kapture_dirpath, image_filename)
        for image_filename in kapture_data
    }


def image_ids_from_feature_dirpath(
        kapture_type: Type[Union[kapture.Keypoints,
                                 kapture.Descriptors,
                                 kapture.GlobalFeatures,
                                 kapture.Matches]],
        kapture_dirpath: str = '') -> Dict[str, str]:
    """
    Populate feature files and returns their corresponding image_filename.

    :param kapture_type:
    :param kapture_dirpath:
    :return: image file path
    """
    feature_dirpath = get_features_fullpath(kapture_type, kapture_dirpath)
    # filter only files with proper extension and
    feature_filenames = populate_files_in_dirpath(feature_dirpath, FEATURE_FILE_EXTENSION[kapture_type])
    # remove file extensions to retrieve image name.
    image_filenames = (feature_filename[:-len(FEATURE_FILE_EXTENSION[kapture_type])]
                       for feature_filename in feature_filenames)
    return image_filenames


def features_check_dir(
        kapture_data: Union[kapture.Keypoints,
                            kapture.Descriptors,
                            kapture.GlobalFeatures,
                            kapture.Matches],
        kapture_dirpath: str) -> bool:
    """
    Makes sure all files actually exists.

    :param kapture_data:
    :param kapture_dirpath:
    :return: True only if all exist, false otherwise
    """
    file_list = features_to_filepaths(kapture_data, kapture_dirpath).values()
    all_files_exists = all(path.exists(feature_filepath) for feature_filepath in file_list)
    return all_files_exists


# image_keypoints ######################################################################################################
def image_keypoints_from_file(filepath: str, dtype: Type, dsize: int) -> np.array:
    """
    Read the image keypoints

    :param filepath: path to the file
    :param dtype: data type
    :param dsize: number of data per keypoint
    :return: the image keypoints
    """
    return array_from_file(filepath, dtype, dsize)


def image_keypoints_to_file(filepath: str, image_keypoints: np.array) -> None:
    """
    Writes the image keypoints to file

    :param filepath: file path
    :param image_keypoints: image keypoints
    """
    array_to_file(filepath, image_keypoints)


def get_keypoints_fullpath(kapture_dirpath: str, image_filename: Optional[str] = None) -> str:
    """
    Computes the full path of the keypoints file

    :param kapture_dirpath: top kapture directory path
    :param image_filename: image file name
    :return: full path of the keypoints file
    """
    return get_features_fullpath(kapture.Keypoints, kapture_dirpath, image_filename)


def keypoints_to_filepaths(keypoints: kapture.Keypoints, kapture_dirpath: str) -> Dict[str, str]:
    """
    Computes keypoints files paths

    :param keypoints: keypoints
    :param kapture_dirpath: top kapture directory path
    :return: keypoint to keypoint file dictionary
    """
    return features_to_filepaths(keypoints, kapture_dirpath)


def keypoints_check_dir(keypoints: kapture.Keypoints, kapture_dirpath: str) -> bool:
    """
    Checks that all keypoints file exist.

    :param keypoints: keypoints
    :param kapture_dirpath: top kapture directory path
    :return: True if they all exist, false otherwise.
    """
    return features_check_dir(keypoints, kapture_dirpath)


# image_descriptors ####################################################################################################
def image_descriptors_from_file(filepath: str, dtype: Type, dsize: int) -> np.array:
    """
    Read the image descriptors

    :param filepath: path to the file
    :param dtype: data type
    :param dsize: number of data per keypoint
    :return: the image descriptors
    """
    return array_from_file(filepath, dtype, dsize)


def image_descriptors_to_file(filepath: str, image_descriptors: np.array) -> None:
    """
    Writes the image descriptors to file

    :param filepath: file path
    :param image_descriptors: image descriptors
    """
    array_to_file(filepath, image_descriptors)


def get_descriptors_fullpath(kapture_dirpath: str, image_filename: Optional[str] = None) -> str:
    """
    Computes the full path of the descriptors file

    :param kapture_dirpath: top kapture directory path
    :param image_filename: image file name
    :return: full path of the descriptors file
    """
    return get_features_fullpath(kapture.Descriptors, kapture_dirpath, image_filename)


def descriptors_to_filepaths(descriptors: kapture.Descriptors, kapture_dirpath: str) -> Dict[str, str]:
    """
    Computes descriptors files paths

    :param descriptors: descriptors
    :param kapture_dirpath: top kapture directory path
    :return: descriptors to descriptors file dictionary
    """
    return features_to_filepaths(descriptors, kapture_dirpath)


def descriptors_check_dir(descriptors: kapture.Descriptors, kapture_dirpath: str) -> bool:
    """
    Checks that all descriptors file exist.

    :param descriptors: descriptors
    :param kapture_dirpath: top kapture directory path
    :return: True if they all exist, false otherwise.
    """
    return features_check_dir(descriptors, kapture_dirpath)


# global_features ######################################################################################################
def image_global_features_from_file(filepath: str, dtype: Type, dsize: int) -> np.array:
    """
    Read the image global features

    :param filepath: path to the file
    :param dtype: data type
    :param dsize: number of data per keypoint
    :return: the global features
    """
    return array_from_file(filepath, dtype, dsize)


def image_global_features_to_file(filepath: str, image_global_descriptor: np.array) -> None:
    """
    Writes the image global features to file

    :param filepath: file path
    :param image_global_descriptor: image global features
    """
    array_to_file(filepath, image_global_descriptor)


def get_global_features_fullpath(kapture_dirpath: str, image_filename: Optional[str] = None) -> str:
    """
    Computes the full path of the global features file

    :param kapture_dirpath: top kapture directory path
    :param image_filename: image file name
    :return: full path of the global features file
    """
    return get_features_fullpath(kapture.GlobalFeatures, kapture_dirpath, image_filename)


def global_features_to_filepaths(global_features: kapture.GlobalFeatures, kapture_dirpath: str) -> Dict[str, str]:
    """
    Computes global features files paths

    :param global_features: global features
    :param kapture_dirpath: top kapture directory path
    :return: global features to global features file dictionary
    """
    return features_to_filepaths(global_features, kapture_dirpath)


def global_features_check_dir(global_features: kapture.GlobalFeatures, kapture_dirpath: str) -> bool:
    """
    Checks that all keypoints file exist.

    :param global_features: global features
    :param kapture_dirpath: top kapture directory path
    :return: True if they all exist, false otherwise.
    """
    return features_check_dir(global_features, kapture_dirpath)


# matches ##############################################################################################################
def image_matches_from_file(filepath: str) -> np.array:
    """
    Read the image matches
     [keypoint_id_image1, keypoint_id_image2, score]

    :param filepath: path to the file
    :return: matches
    """
    return array_from_file(filepath, dtype=np.float64, dsize=3)


def image_matches_to_file(filepath: str, image_matches: np.array) -> None:
    """
    Writes the image matches to file

    :param filepath: file path
    :param image_matches: image matches
    """

    assert image_matches.dtype == np.float64
    assert image_matches.shape[1] == 3
    array_to_file(filepath, image_matches)


def get_matches_fullpath(image_filename_pair: Optional[Tuple[str, str]] = None, kapture_dirpath: str = '') -> str:
    """
    Computes the full path of the matches file between two images

    :param kapture_dirpath: top kapture directory path
    :param image_filename_pair: image file names
    :return: full path of the matches file
    """

    if image_filename_pair is not None:
        filename = path.join(image_filename_pair[0] + FEATURE_PAIR_PATH_SEPARATOR[kapture.Matches],
                             image_filename_pair[1])
    else:
        filename = None
    return get_features_fullpath(kapture.Matches, kapture_dirpath, filename)


def matches_to_filepaths(matches: kapture.Matches, kapture_dirpath: str = '') -> Dict[Tuple[str, str], str]:
    """
    Computes matches files paths

    :param matches: matches
    :param kapture_dirpath: top kapture directory path
    :return: file to file matches to matches file dictionary
    """
    return {
        image_filename_pair: get_matches_fullpath(image_filename_pair, kapture_dirpath)
        for image_filename_pair in matches
    }


def matching_pairs_from_dirpath(kapture_dirpath: str) -> Iterable[Tuple[str, str]]:
    """
    Read and build Matches from kapture directory tree.
    """
    matches_dirpath = get_matches_fullpath(None, kapture_dirpath)
    # list all files there is
    # filter only match files (the ones endings with .matches)
    matches_filenames = populate_files_in_dirpath(matches_dirpath, FEATURE_FILE_EXTENSION[kapture.Matches])

    # remove the extensions and cut
    matching_pairs = (
        path_secure(matches_filename)[:-len(FEATURE_FILE_EXTENSION[kapture.Matches])
                                      ].split(FEATURE_PAIR_PATH_SEPARATOR[kapture.Matches] + '/')
        for matches_filename in matches_filenames
    )
    matching_pairs = ((matches[0], matches[1]) for matches in matching_pairs if len(matches) == 2)
    return matching_pairs


def matches_check_dir(matches: kapture.Matches, kapture_dirpath: str) -> bool:
    """
    Checks that all matches file exist.

    :param matches: matches
    :param kapture_dirpath: top kapture directory path
    :return: True if they all exist, false otherwise.
    """
    file_list = (get_matches_fullpath(pair, kapture_dirpath) for pair in matches)
    all_files_exists = all(path.exists(filepath) for filepath in file_list)
    return all_files_exists



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
All reading and writing operations of kapture objects in CSV like files
"""

import os
import os.path as path
import re
from typing import Any, List, Optional, Set, Type, Union
from collections import namedtuple
import numpy as np

import kapture
import kapture.io.features

logger = kapture.logger

# file names conventions
CSV_FILENAMES = {
    kapture.Sensors: path.join('sensors', 'sensors.txt'),
    kapture.Trajectories: path.join('sensors', 'trajectories.txt'),
    kapture.Rigs: path.join('sensors', 'rigs.txt'),
    kapture.RecordsCamera: path.join('sensors', 'records_camera.txt'),
    kapture.RecordsDepth: path.join('sensors', 'records_depth.txt'),
    kapture.RecordsLidar: path.join('sensors', 'records_lidar.txt'),
    kapture.RecordsWifi: path.join('sensors', 'records_wifi.txt'),
    kapture.RecordsBluetooth: path.join('sensors', 'records_bluetooth.txt'),
    kapture.RecordsGnss: path.join('sensors', 'records_gnss.txt'),
    kapture.RecordsAccelerometer: path.join('sensors', 'records_accelerometer.txt'),
    kapture.RecordsGyroscope: path.join('sensors', 'records_gyroscope.txt'),
    kapture.RecordsMagnetic: path.join('sensors', 'records_magnetic.txt'),
    kapture.Points3d: path.join('reconstruction', 'points3d.txt'),
    kapture.Keypoints: path.join('reconstruction', 'keypoints', 'keypoints.txt'),
    kapture.Descriptors: path.join('reconstruction', 'descriptors', 'descriptors.txt'),
    kapture.GlobalFeatures: path.join('reconstruction', 'global_features', 'global_features.txt'),
    kapture.Observations: path.join('reconstruction', 'observations.txt'),
}


def get_csv_fullpath(kapture_type: Any, kapture_dirpath: str = '') -> str:
    """
    Returns the full path to csv kapture file for a given datastructure and root directory.
    This path is the concatenation of the kapture root path and subpath into kapture into data structure.

    :param kapture_type: type of kapture data (kapture.RecordsCamera, kapture.Trajectories, ...)
    :param kapture_dirpath: root kapture path
    :return: full path of csv file for that type of data
    """
    filename = CSV_FILENAMES[kapture_type]
    return path.join(kapture_dirpath, filename)


PADDINGS = {
    'timestamp': [8],
    'device_id': [3],
    'pose': [4, 4, 4, 4, 4, 4, 4],
}

KAPTURE_FORMAT_1 = "# kapture format: 1.0"
KAPTURE_FORMAT_PARSING_RE = '# kapture format\\:\\s*(?P<version>\\d+\\.\\d+)'


def get_version_from_header(header_string: str) -> Optional[str]:
    """
    Get the kapture format version from the given header string.

    :param header_string: a header string
    :return: version as string (i.e. '2.1') if found, None otherwise
    """
    m = re.search(KAPTURE_FORMAT_PARSING_RE, header_string)
    if m:
        return m['version']
    return None


def get_version_from_csv_file(csv_file_path: str) -> Optional[str]:
    """
    Get the kapture format version from a file.

    :param csv_file_path: path to the csv file
    :return: version as string (i.e. '2.1') if found, None otherwise
    """
    if path.isfile(csv_file_path):
        with open(csv_file_path) as f:
            first_line = f.readline()
            return get_version_from_header(first_line)
    return None


def current_format_version() -> str:
    """
    Get the current format version

    :return: format version
    """
    return get_version_from_header(KAPTURE_FORMAT_1)


def kapture_format_version(kapture_dirpath: str) -> Optional[str]:
    """
    Reads kapture format version.

    :param kapture_dirpath: kapture directory root path
    :return: kapture format version if found.
    """
    sensors_file_path = path.join(kapture_dirpath, CSV_FILENAMES[kapture.Sensors])
    return get_version_from_csv_file(sensors_file_path)


def float_safe(representation) -> Union[float, None]:
    """
    Safe float cast
    https://stackoverflow.com/questions/6330071/safe-casting-in-python

    :param representation: to cast
    :return: float value if the value was castable, or None
    """
    try:
        return float(representation)
    except (ValueError, TypeError):
        return None


def float_array_or_none(representation_list) -> Union[List[float], None]:
    """
    Safe cast of list of float representations
    https://stackoverflow.com/questions/6330071/safe-casting-in-python

    :param representation_list: list of values to convert
    :return: an array of floats or None if a single one is invalid
    """
    array = [float_safe(v) for v in representation_list]
    return array if not any(v is None for v in array) else None


def table_to_file(file, table, header=None, padding=None) -> None:
    """
    Writes the given table (list of list) into a file.
            The file must be previously open as write mode.
            If table is a generator, must be valid at runtime.

    :param file: file id opened in write mode (with open(filepath, 'w') as file:)
    :param table: an iterable of iterable
    :param header: row added at the beginning of the file (+\n)
    :param padding: the padding of each column as a list of int of same size of the rows of the table.
    :return:
    """
    if header:
        file.write(KAPTURE_FORMAT_1 + '\n')
        file.write(header + '\n')
    for row in table:
        if padding:
            row = [str(v).rjust(padding[i]) for i, v in enumerate(row)]
        file.write(', '.join(f'{v}' for v in row) + '\n')


def table_from_file(file):
    """
    Returns an iterable of iterable (generator) on the opened file.
        Be aware that the returned generator is valid as long as file is valid.

    :param file: file id opened in read mode (with open(filepath, 'r') as file:)
    :return: an iterable of iterable on kapture objects values
    """
    table = file.readlines()
    # remove comment lines
    table = (l1 for l1 in table if not l1.startswith('#'))
    # remove empty lines
    table = (l2 for l2 in table if l2.strip())
    # trim trailing EOL
    table = (l3.rstrip("\n\r") for l3 in table)
    # split comma (and trim afterwards spaces)
    table = (re.split(r'\s*,\s*', l4) for l4 in table)
    return table


########################################################################################################################
# poses  ###############################################################################################################
def pose_to_list(pose: kapture.PoseTransform) -> List[Union[float, str]]:
    """
    Converts a 6D pose to a list of floats to save in CSV

    :param pose: 6D pose
    :return: list of float
    """
    assert (isinstance(pose, kapture.PoseTransform))
    rotation = pose.r_raw if pose.r is not None else 4 * ['']
    translation = pose.t_raw if pose.t is not None else 3 * ['']
    return rotation + translation


########################################################################################################################
# Sensor ###############################################################################################################
def sensor_to_list(sensor: kapture.Sensor) -> List[str]:
    """
    Converts a sensor into a list of strings to save in CSV

    :param sensor:
    :return: list of strings
    """
    assert (isinstance(sensor, kapture.Sensor))
    # sensor_id, name, model, [model_params]+
    fields = [sensor.name or ''] + [sensor.sensor_type] + [str(v) for v in sensor.sensor_params]
    return fields


########################################################################################################################
# Sensors ##############################################################################################################
def sensors_to_file(filepath: str, sensors: kapture.Sensors) -> None:
    """
    Writes the sensors to CSV file.

    :param filepath: input file path
    :param sensors: input sensors
    """
    assert (isinstance(sensors, kapture.Sensors))
    header = '# sensor_id, name, sensor_type, [sensor_params]+'
    table = ([sensor_id] + sensor_to_list(sensor)
             for sensor_id, sensor in sensors.items())

    os.makedirs(path.dirname(filepath), exist_ok=True)
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header)


def sensors_from_file(filepath: str) -> kapture.Sensors:
    """
    Reads sensors from CSV file.

    :param filepath: input file path
    :return: sensors
    """
    sensors = kapture.Sensors()
    with open(filepath) as file:
        table = table_from_file(file)
        # sensor_id, name, sensor_type, [sensor_params]+'
        for sensor_id, name, sensor_type, *sensor_params in table:
            sensor = kapture.create_sensor(sensor_type=sensor_type, sensor_params=sensor_params, name=name)
            sensors[sensor_id] = sensor

    return sensors


########################################################################################################################
# Rig ##################################################################################################################
def rigs_to_file(filepath: str, rigs: kapture.Rigs) -> None:
    """
    Writes rigs to CSV file.

    :param filepath:
    :param rigs:
    """
    assert (isinstance(rigs, kapture.Rigs))
    header = '# rig_id, sensor_id, qw, qx, qy, qz, tx, ty, tz'
    padding = PADDINGS['device_id'] + PADDINGS['device_id'] + PADDINGS['pose']
    table = ([rig_id, sensor_id] + pose_to_list(pose)
             for rig_id, rig in rigs.items()
             for sensor_id, pose in rig.items())

    os.makedirs(path.dirname(filepath), exist_ok=True)
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header, padding=padding)


def rigs_from_file(filepath: str, sensor_ids: Optional[Set[str]] = None) -> kapture.Rigs:
    """
    Reads rigs from CSV file.

    :param filepath: input file path
    :param sensor_ids: input set of valid sensor ids.
                        If a rig id collides one of them, raise error.
                        If a sensor in rig is not in sensor_ids, it is ignored.
    :return: rigs
    """
    # rig_id, sensor_id, qw, qx, qy, qz, tx, ty, tz
    rigs = kapture.Rigs()
    with open(filepath) as file:
        table = table_from_file(file)
        for rig_id, sensor_id, qw, qx, qy, qz, tx, ty, tz in table:
            if rig_id in sensor_ids:
                raise ValueError(f'collision between a sensor ID and rig ID ({rig_id})')
            rotation = float_array_or_none([qw, qx, qy, qz])
            translation = float_array_or_none([tx, ty, tz])
            pose = kapture.PoseTransform(rotation, translation)
            rigs[str(rig_id), sensor_id] = pose

    if sensor_ids is not None:
        # expunge all undesired sensors
        rig_ids = set(rigs)
        for rig_id in rig_ids:
            for sensor_id in set(rigs[rig_id]):
                if sensor_id not in sensor_ids and sensor_id not in rig_ids:
                    logger.debug(f'dropping sensor {sensor_id} from rig {rig_id} because it is unknown sensor.')
                    del rigs[rig_id][sensor_id]

    return rigs


########################################################################################################################
# Trajectories #########################################################################################################
def trajectories_to_file(filepath: str, trajectories: kapture.Trajectories) -> None:
    """
    Writes trajectories to CSV file.

    :param filepath:
    :param trajectories:
    """
    assert (isinstance(trajectories, kapture.Trajectories))
    header = '# timestamp, device_id, qw, qx, qy, qz, tx, ty, tz'
    padding = PADDINGS['timestamp'] + PADDINGS['device_id'] + PADDINGS['pose']
    table = (
        [timestamp, sensor_id] + pose_to_list(trajectories[(timestamp, sensor_id)])
        for timestamp, sensor_id in sorted(trajectories.key_pairs())
    )

    os.makedirs(path.dirname(filepath), exist_ok=True)
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header, padding=padding)


def trajectories_from_file(filepath: str, device_ids: Optional[Set[str]] = None) -> kapture.Trajectories:
    """
    Reads trajectories from CSV file.

    :param filepath: input file path
    :param device_ids: input set of valid device ids (rig or sensor).
                        If the trajectories contains unknown devices, they will be ignored.
                        If no device_ids given, everything is loaded.
    :return: trajectories
    """
    trajectories = kapture.Trajectories()
    with open(filepath) as file:
        table = table_from_file(file)
        # timestamp, device_id, qw, qx, qy, qz, tx, ty, tz
        for timestamp, device_id, qw, qx, qy, qz, tx, ty, tz in table:
            if device_ids is not None and device_id not in device_ids:
                # just ignore
                continue
            rotation = float_array_or_none([qw, qx, qy, qz])
            trans = float_array_or_none([tx, ty, tz])
            pose = kapture.PoseTransform(rotation, trans)
            trajectories[(int(timestamp), str(device_id))] = pose
    return trajectories


########################################################################################################################
# Records Camera #######################################################################################################
def records_camera_to_file(filepath: str, records_camera: kapture.RecordsCamera) -> None:
    """
    Writes records_camera to CSV file.

    :param filepath:
    :param records_camera:
    """
    assert (isinstance(records_camera, kapture.RecordsCamera))
    header = '# timestamp, device_id, image_path'
    table = (
        [timestamp, sensor_id] + [records_camera[(timestamp, sensor_id)]]
        for timestamp, sensor_id in sorted(records_camera.key_pairs())
    )
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header)


def records_camera_from_file(filepath: str, camera_ids: Optional[Set[str]] = None) -> kapture.RecordsCamera:
    """
    Reads records_camera from CSV file.

    :param filepath: input file path
    :param camera_ids: input set of valid camera device ids.
                        If the records_camera contains unknown devices, they will be ignored.
                        If not given, all cameras are loaded.
    :return: camera records
    """
    records_camera = kapture.RecordsCamera()
    with open(filepath) as file:
        table = table_from_file(file)
        # timestamp, device_id, image_path
        for timestamp, device_id, image_path in table:
            if camera_ids is not None and device_id not in camera_ids:
                # just ignore
                continue
            records_camera[(int(timestamp), str(device_id))] = image_path
    return records_camera


########################################################################################################################
# Records Depth #######################################################################################################
def records_depth_to_file(filepath: str, records_depth: kapture.RecordsDepth) -> None:
    """
    Writes records_depth to CSV file.

    :param filepath:
    :param records_depth:
    """
    assert (isinstance(records_depth, kapture.RecordsDepth))
    header = '# timestamp, device_id, depth_map_path'
    table = (
        [timestamp, sensor_id] + [records_depth[(timestamp, sensor_id)]]
        for timestamp, sensor_id in sorted(records_depth.key_pairs())
    )
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header)


def records_depth_from_file(filepath: str, camera_ids: Optional[Set[str]] = None) -> kapture.RecordsDepth:
    """
    Reads records_depth from CSV file.

    :param filepath: input file path
    :param camera_ids: input set of valid camera device ids.
                        If the records_camera contains unknown devices, they will be ignored.
                        If not given, all cameras are loaded.
    :return: camera records
    """
    records_depth = kapture.RecordsDepth()
    with open(filepath) as file:
        table = table_from_file(file)
        # timestamp, device_id, image_path
        for timestamp, device_id, depth_map_path in table:
            if camera_ids is not None and device_id not in camera_ids:
                # just ignore
                continue
            records_depth[(int(timestamp), str(device_id))] = depth_map_path
    return records_depth


########################################################################################################################
# Records Lidar ########################################################################################################
def records_lidar_to_file(filepath: str, records_lidar: kapture.RecordsLidar) -> None:
    """
    Writes records_lidar to CSV file.

    :param filepath:
    :param records_lidar:
    """
    assert (isinstance(records_lidar, kapture.RecordsLidar))
    header = '# timestamp, device_id, point_cloud_path'
    table = (
        [timestamp, sensor_id] + [records_lidar[(timestamp, sensor_id)]]
        for timestamp, sensor_id in sorted(records_lidar.key_pairs())
    )
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header)


def records_lidar_from_file(
        filepath: str,
        lidar_ids: Optional[Set[str]] = None
) -> kapture.RecordsLidar:
    """
    Reads records_lidar from CSV file.

    :param filepath: input file path
    :param lidar_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                            If omitted, then it loads all devices.
    :return: Lidar records
    """
    records_lidar = kapture.RecordsLidar()
    with open(filepath) as file:
        table = table_from_file(file)
        # timestamp, device_id, point_cloud_path
        for timestamp, device_id, point_cloud_path in table:
            if lidar_ids is not None and device_id not in lidar_ids:
                # just ignore
                continue
            records_lidar[(int(timestamp), str(device_id))] = point_cloud_path
    return records_lidar


########################################################################################################################
def records_generic_to_file(
        filepath: str,
        records: kapture.RecordsBase) -> None:
    """
        Writes records_wifi to file

        :param filepath: path where to save records file.
        :param records:
        """
    assert (isinstance(records, kapture.RecordsBase))
    header = '# timestamp, device_id, ' + ', '.join(f.name for f in records.record_type.fields())
    table = []
    for timestamp, sensor_id, record in kapture.flatten(records, is_sorted=True):
        table.append([timestamp, sensor_id] + [str(v) for v in record.astuple()])
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header)


def records_generic_from_file(
        records_type: Type,
        filepath: str,
        sensor_ids: Optional[Set[str]] = None
) -> Union[kapture.RecordsBase,
           kapture.RecordsGnss,
           kapture.RecordsGyroscope,
           kapture.RecordsAccelerometer,
           kapture.RecordsMagnetic]:
    """
    Reads Records data from CSV file.

    :param records_type: type of records expected (eg RecordsWifi)
    :param filepath: input file path
    :param sensor_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                     If omitted, then it loads all devices.
    :return: records
    """
    records = records_type()
    with open(filepath) as file:
        table = table_from_file(file)
        # timestamp, device_id, *
        for timestamp, device_id, *data in table:
            timestamp = int(timestamp)
            device_id = str(device_id)
            if sensor_ids is not None and device_id not in sensor_ids:
                # just ignore
                continue
            records[timestamp, device_id] = records_type.record_type(*data)

    return records


# Records Wifi #########################################################################################################
def records_wifi_to_file(filepath: str, records_wifi: kapture.RecordsWifi) -> None:
    """
    Writes records_wifi to file

    :param filepath:
    :param records_wifi:
    """
    assert (isinstance(records_wifi, kapture.RecordsWifi))
    header = '# timestamp, device_id, BSSID, frequency, RSSI, SSID, scan_time_start, scan_time_end'
    table = []
    for timestamp, sensor_id in sorted(records_wifi.key_pairs()):
        for bssid, record in records_wifi[timestamp, sensor_id].items():
            table.append([timestamp, sensor_id, bssid] + [str(v) for v in record.astuple()])
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header)


def records_wifi_from_file(
        filepath: str,
        sensor_ids: Optional[Set[str]] = None
) -> kapture.RecordsWifi:
    """
    Reads RecordsWifi from CSV file.

    :param filepath: input file path
    :param sensor_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                            If omitted, then it loads all devices.
    :return: Wifi records
    """
    records_wifi = kapture.RecordsWifi()
    with open(filepath) as file:
        table = table_from_file(file)
        # timestamp, device_id, BSSID, frequency, RSSI, SSID, scan_time_start, scan_time_end
        for timestamp, device_id, BSSID, frequency, RSSI, SSID, scan_time_start, scan_time_end in table:
            timestamp, device_id = int(timestamp), str(device_id)
            if sensor_ids is not None and device_id not in sensor_ids:
                # just ignore
                continue
            if (timestamp, device_id) not in records_wifi:
                records_wifi[timestamp, device_id] = kapture.RecordWifi()
            records_wifi[timestamp, device_id][BSSID] = kapture.RecordWifiSignal(
                frequency, RSSI, SSID, scan_time_start, scan_time_end)

    return records_wifi


# Records Bluetooth ####################################################################################################
def records_bluetooth_to_file(
        filepath: str,
        records_bluetooth: kapture.RecordsBluetooth
) -> None:
    """
    Writes records_wifi to file

    :param filepath: output file path.
    :param records_bluetooth:
    """
    assert (isinstance(records_bluetooth, kapture.RecordsBluetooth))
    header = '# timestamp, device_id, address, RSSI, name'
    table = []
    for timestamp, sensor_id in sorted(records_bluetooth.key_pairs()):
        for address, bt_record in records_bluetooth[timestamp, sensor_id].items():
            table.append([timestamp, sensor_id, address] + [str(v) for v in bt_record.astuple()])
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header)


def records_bluetooth_from_file(
        filepath: str,
        sensor_ids: Optional[Set[str]] = None
) -> kapture.RecordsBluetooth:
    """
    Reads RecordsWifi from CSV file.

    :param filepath: input file path
    :param sensor_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                        If omitted, then it loads all devices.
    :return: Wifi records
    """
    records_bluetooth = kapture.RecordsBluetooth()
    with open(filepath) as file:
        table = table_from_file(file)
        # timestamp, device_id, address, RSSI, name
        for timestamp, device_id, address, RSSI, name in table:
            timestamp, device_id = int(timestamp), str(device_id)
            if sensor_ids is not None and device_id not in sensor_ids:
                # just ignore
                continue
            if (timestamp, device_id) not in records_bluetooth:
                records_bluetooth[timestamp, device_id] = kapture.RecordBluetooth()
            records_bluetooth[timestamp, device_id][address] = kapture.RecordBluetoothSignal(
                rssi=RSSI, name=name)

    return records_bluetooth


# GNSS #################################################################################################################
def records_gnss_to_file(
        filepath: str,
        records_gnss: kapture.RecordsGnss
) -> None:
    records_generic_to_file(filepath, records_gnss)


def records_gnss_from_file(
        filepath: str,
        sensor_ids: Optional[Set[str]] = None
) -> kapture.RecordsGnss:
    return records_generic_from_file(
        records_type=kapture.RecordsGnss,
        filepath=filepath,
        sensor_ids=sensor_ids
    )


# Accelerometer ########################################################################################################
def records_accelerometer_to_file(
        filepath: str,
        records_accelerometer: kapture.RecordsAccelerometer
) -> None:
    records_generic_to_file(filepath, records_accelerometer)


def records_accelerometer_from_file(
        filepath: str,
        sensor_ids: Optional[Set[str]] = None
) -> kapture.RecordsAccelerometer:
    return records_generic_from_file(
        records_type=kapture.RecordsAccelerometer,
        filepath=filepath,
        sensor_ids=sensor_ids
    )


# Gyroscope ########################################################################################################
def records_gyroscope_to_file(
        filepath: str,
        records_gyroscope: kapture.RecordsGyroscope
) -> None:
    records_generic_to_file(filepath, records_gyroscope)


def records_gyroscope_from_file(
        filepath: str,
        sensor_ids: Optional[Set[str]] = None
) -> kapture.RecordsGyroscope:
    return records_generic_from_file(
        records_type=kapture.RecordsGyroscope,
        filepath=filepath,
        sensor_ids=sensor_ids
    )


# Magnetic ########################################################################################################
def records_magnetic_to_file(
        filepath: str,
        records_magnetic: kapture.RecordsMagnetic
) -> None:
    records_generic_to_file(filepath, records_magnetic)


def records_magnetic_from_file(
        filepath: str,
        sensor_ids: Optional[Set[str]] = None
) -> kapture.RecordsMagnetic:
    return records_generic_from_file(
        records_type=kapture.RecordsMagnetic,
        filepath=filepath,
        sensor_ids=sensor_ids
    )


########################################################################################################################
# features #############################################################################################################
ImageFeatureConfig = namedtuple('ImageFeatureConfig', ['name', 'dtype', 'dsize'])


# config file ##########################################################################################################
def image_features_config_to_file(config_filepath: str, config: ImageFeatureConfig) -> None:
    """
    Writes feature config files, ie. name, data type, and data size.

    :param config_filepath: input path to config file to write.
    :param config: input config to be written.
    """
    os.makedirs(path.dirname(config_filepath), exist_ok=True)
    header = "# name, dtype, dsize"
    dtype = str(config.dtype) if isinstance(config.dtype, np.dtype) else config.dtype.__name__
    line = [config.name, dtype, str(config.dsize)]
    with open(config_filepath, 'wt') as file:
        table_to_file(file, [line], header=header)


def image_features_config_from_file(config_filepath: str) -> ImageFeatureConfig:
    """
    Reads feature config files, ie. name, data type, and data size.

    :param config_filepath: input path to config file to read.
    :return: config The read config.
    """
    if not path.exists(config_filepath):
        raise FileNotFoundError(f'{path.basename(config_filepath)} file is missing')

    with open(config_filepath, 'rt') as file:
        table = table_from_file(file)
        line = list(table)[0]
        assert len(line) == 3
        name, dtype, dsize = line[0], line[1], int(line[2])

    # try to list all possible type from numpy that can be used in eval(dtype)
    from numpy import float, float32, float64, int32, uint8  # noqa: F401
    if isinstance(type(eval(dtype)), type):
        dtype = eval(dtype)
    else:
        raise ValueError('Expect data type ')
    return ImageFeatureConfig(name, dtype, dsize)


# files #########################################################################################################
def image_feature_to_file(
        config_filepath: str,
        image_features: Union[kapture.Keypoints, kapture.Descriptors, kapture.GlobalFeatures]
) -> None:
    """
    Writes ImageFeatures config file only.

    :param config_filepath: eg. /path/to/keypoints/keypoints.txt
    :param image_features: input feature date.
    """
    config = ImageFeatureConfig(image_features.type_name, image_features.dtype, image_features.dsize)
    image_features_config_to_file(config_filepath, config)


def image_features_from_dir(
        kapture_type: Type[Union[kapture.Keypoints,
                                 kapture.Descriptors,
                                 kapture.GlobalFeatures]],
        kapture_dirpath: str,
        image_filenames: Optional[Set[str]]
) -> kapture.ImageFeatures:
    """
    Reads and builds ImageFeatures from images_filenames if given, or directly from actual files in  root_dirpath.

    :param kapture_type: kapture class type.
    :param kapture_dirpath: input path to kapture root directory.
    :param image_filenames: None or set of image relative paths.
    :return: Features
    """

    # make config_filepath from data_dirpath
    config_filepath = get_csv_fullpath(kapture_type, kapture_dirpath)
    config = image_features_config_from_file(config_filepath)
    if image_filenames is None:
        # images_path is empty, so populates all feature files
        image_filenames_generator = kapture.io.features.image_ids_from_feature_dirpath(kapture_type, kapture_dirpath)
    else:
        # filter only existing files
        image_filenames_generator = (
            image_name
            for image_name in image_filenames
            if path.exists(kapture.io.features.get_features_fullpath(kapture_type, kapture_dirpath, image_name)))

    image_filenames = set(image_filenames_generator)
    return kapture_type(config.name, config.dtype, config.dsize, image_filenames)


########################################################################################################################
# keypoints ############################################################################################################
def keypoints_to_file(config_filepath: str, keypoints: kapture.Keypoints) -> None:
    """
    Writes keypoints to CSV file.

    :param config_filepath:
    :param keypoints:
    """
    return image_feature_to_file(config_filepath=config_filepath,
                                 image_features=keypoints)


def keypoints_from_dir(kapture_dirpath: str, images_paths: Optional[Set[str]]) -> kapture.Keypoints:
    """
    Reads and builds keypoints from images_filenames if given, or directly from actual files in kapture_dirpath.

    :param kapture_dirpath: root path of kapture
    :param images_paths: optional list of image file names
    :return: Keypoints
    """
    return image_features_from_dir(kapture_type=kapture.Keypoints,
                                   kapture_dirpath=kapture_dirpath,
                                   image_filenames=images_paths)


########################################################################################################################
# descriptors ##########################################################################################################
def descriptors_to_file(config_filepath: str, descriptors: kapture.Descriptors) -> None:
    """
    Writes descriptors to CSV file.

    :param config_filepath:
    :param descriptors:
    """
    return image_feature_to_file(config_filepath=config_filepath,
                                 image_features=descriptors)


def descriptors_from_dir(kapture_dirpath: str, images_paths: Set[str]) -> kapture.Descriptors:
    """
    Reads and builds descriptors from images_filenames if given, or directly from actual files in kapture_dirpath.

    :param kapture_dirpath: root path of kapture
    :param images_paths: optional list of image file names
    :return: Descriptors
    """
    return image_features_from_dir(kapture_type=kapture.Descriptors,
                                   kapture_dirpath=kapture_dirpath,
                                   image_filenames=images_paths)


########################################################################################################################
# global_features ######################################################################################################
def global_features_to_file(config_filepath: str, global_features: kapture.GlobalFeatures) -> None:
    """
    Writes global features to CSV file.

    :param config_filepath:
    :param global_features:
    """
    return image_feature_to_file(config_filepath=config_filepath,
                                 image_features=global_features)


def global_features_from_dir(kapture_dirpath: str, images_paths: Set[str]) -> kapture.GlobalFeatures:
    """
    Reads and builds Global features from images_filenames if given, or directly from actual files in kapture_dirpath.

    :param kapture_dirpath: root path of kapture
    :param images_paths: optional list of image file names
    :return: Global features
    """

    return image_features_from_dir(kapture_type=kapture.GlobalFeatures,
                                   kapture_dirpath=kapture_dirpath,
                                   image_filenames=images_paths)


########################################################################################################################
# matches ##############################################################################################################
def matches_from_dir(kapture_dirpath: str,
                     image_filenames: Optional[Set[str]] = None,
                     matches_pairsfile_path: Optional[str] = None) -> kapture.Matches:
    """
    Reads and builds Matches from images_filenames if given, or directly from actual files in kapture_dirpath.

    :param kapture_dirpath: root path of kapture
    :param image_filenames: optional list of image file names
    :param matches_pairsfile_path: text file in the csv format; where each line is image_name1, image_name2, score
    :return: Matches
    """

    if matches_pairsfile_path is None:
        # populate files from disk
        match_pairs_generator = kapture.io.features.matching_pairs_from_dirpath(kapture_dirpath)
    else:
        with open(matches_pairsfile_path, 'r') as fid:
            table = table_from_file(fid)
            # get matches list from pairsfile
            match_pairs_generator = ((query_name, map_name) if query_name < map_name else (map_name, query_name)
                                     for query_name, map_name, _ in table)
            # keeps only the one that actually exists on disk
            match_pairs_generator = (image_pair
                                     for image_pair in match_pairs_generator
                                     if path.isfile(kapture.io.features.get_matches_fullpath(image_pair,
                                                                                             kapture_dirpath))
                                     )

    if image_filenames is not None:
        # retains only files that correspond to known images
        match_pairs_generator = (
            image_pair
            for image_pair in match_pairs_generator
            if image_pair[0] in image_filenames and image_pair[1] in image_filenames
        )
    match_pairs = set(match_pairs_generator)
    return kapture.Matches(match_pairs)


########################################################################################################################
# points3d #############################################################################################################
def points3d_to_file(filepath: str, points3d: kapture.Points3d) -> None:
    """
    Writes 3d points to CSV file.

    :param filepath: path to CSV file
    :param points3d: the 3d points
    """
    assert isinstance(points3d, kapture.Points3d)
    os.makedirs(path.dirname(filepath), exist_ok=True)
    header = 'X, Y, Z, R, G, B'
    np.savetxt(filepath, points3d.as_array(), delimiter=',', header=header)


def points3d_from_file(filepath: str) -> kapture.Points3d:
    """
    Reads 3d points from CSV file.

    :param filepath: path to CSV file
    :return: the 3d points
    """

    data = np.loadtxt(filepath, dtype=np.float, delimiter=',', comments='#')
    data = data.reshape((-1, 6))  # make sure of the shape, even if single line file.
    return kapture.Points3d(data)


########################################################################################################################
# observations #########################################################################################################
def observations_to_file(observations_filepath: str, observations: kapture.Observations) -> None:
    """
    Writes observations to CSV file.

    :param observations_filepath: input path to CSV file of observation to write.
                                    Containing directory is created if needed.
    :param observations: input observations to be written.
    """
    assert path.basename(observations_filepath) == 'observations.txt'
    assert isinstance(observations, kapture.Observations)
    header = '# point3d_id, [image_path, feature_id]*'
    table = (
        [str(point3d_idx)] + [str(k) for pair in observations[point3d_idx] for k in pair]
        for point3d_idx in sorted(observations.keys())
    )
    os.makedirs(path.dirname(observations_filepath), exist_ok=True)
    with open(observations_filepath, 'w') as file:
        table_to_file(file, table, header=header)


def observations_from_file(
        observations_filepath: str,
        images_paths_with_keypoints: Optional[Set[str]] = None
) -> kapture.Observations:
    """
    Reads observations from CSV file.

    :param observations_filepath: path to CSV file to read.
    :param images_paths_with_keypoints: input set of image names (ids) that have keypoints.
                                        If given, used to filter out irrelevant observations.
                                        You can get from set(kapture.keypoints)
    :return: observations
    """
    assert path.basename(observations_filepath) == 'observations.txt'
    assert images_paths_with_keypoints is None \
        or (isinstance(images_paths_with_keypoints, set) and len(images_paths_with_keypoints) > 0)
    observations = kapture.Observations()
    with open(observations_filepath) as file:
        table = table_from_file(file)
        # point3d_id, [image_path, feature_id]*
        for points3d_id_str, *pairs in table:
            points3d_id = int(points3d_id_str)
            if len(pairs) > 1:
                image_paths = pairs[0::2]
                keypoints_ids = pairs[1::2]
                for image_path, keypoint_id in zip(image_paths, keypoints_ids):
                    if images_paths_with_keypoints is not None and image_path not in images_paths_with_keypoints:
                        # image_path does not exist in kapture (perhaps it was removed), ignore it
                        continue
                    observations.add(points3d_id, image_path, int(keypoint_id))
    return observations


########################################################################################################################
# Kapture Write ########################################################################################################
KAPTURE_ATTRIBUTE_WRITERS = {
    kapture.Sensors: sensors_to_file,
    kapture.Rigs: rigs_to_file,
    kapture.Trajectories: trajectories_to_file,
    kapture.RecordsCamera: records_camera_to_file,
    kapture.RecordsDepth: records_depth_to_file,
    kapture.RecordsLidar: records_lidar_to_file,
    kapture.RecordsWifi: records_wifi_to_file,
    kapture.RecordsBluetooth: records_bluetooth_to_file,
    kapture.RecordsGnss: records_gnss_to_file,
    kapture.RecordsAccelerometer: records_accelerometer_to_file,
    kapture.RecordsGyroscope: records_gyroscope_to_file,
    kapture.RecordsMagnetic: records_magnetic_to_file,
    kapture.Keypoints: keypoints_to_file,
    kapture.Descriptors: descriptors_to_file,
    kapture.GlobalFeatures: global_features_to_file,
    kapture.Points3d: points3d_to_file,
    kapture.Observations: observations_to_file,
}

KAPTURE_ATTRIBUTE_NAMES = {  # used to list attributes to be saved
    kapture.Sensors: 'sensors',
    kapture.Rigs: 'rigs',
    kapture.Trajectories: 'trajectories',
    kapture.RecordsCamera: 'records_camera',
    kapture.RecordsDepth: 'records_depth',
    kapture.RecordsLidar: 'records_lidar',
    kapture.RecordsWifi: 'records_wifi',
    kapture.RecordsBluetooth: 'records_bluetooth',
    kapture.RecordsGnss: 'records_gnss',
    kapture.RecordsAccelerometer: 'records_accelerometer',
    kapture.RecordsGyroscope: 'records_gyroscope',
    kapture.RecordsMagnetic: 'records_magnetic',
    kapture.Keypoints: 'keypoints',
    kapture.Descriptors: 'descriptors',
    kapture.GlobalFeatures: 'global_features',
    kapture.Points3d: 'points3d',
    kapture.Observations: 'observations',
}


def kapture_to_dir(dirpath: str, kapture_data: kapture.Kapture) -> None:
    """
    Saves kapture data to given directory.

    :param dirpath: input directory root path
    :param kapture_data: input kapture data
    """
    kapture_subtype_to_filepaths = {kapture_class: path.join(dirpath, filename)
                                    for kapture_class, filename in CSV_FILENAMES.items()}
    # save each member of kapture data
    for kapture_class, kapture_member_name in KAPTURE_ATTRIBUTE_NAMES.items():
        part_data = kapture_data.__getattribute__(kapture_member_name)
        if part_data is not None:
            # save it
            logger.debug(f'saving {kapture_member_name} ...')
            write_function = KAPTURE_ATTRIBUTE_WRITERS[kapture_class]
            write_function(kapture_subtype_to_filepaths[kapture_class], part_data)


# Kapture Read #########################################################################################################
# list all data members of kapture.
KAPTURE_LOADABLE_TYPES = {
    kapture.Sensors,
    kapture.Rigs,
    kapture.Trajectories,
    kapture.RecordsCamera,
    kapture.RecordsDepth,
    kapture.RecordsLidar,
    kapture.RecordsWifi,
    kapture.RecordsBluetooth,
    kapture.RecordsGnss,
    kapture.RecordsAccelerometer,
    kapture.RecordsGyroscope,
    kapture.RecordsMagnetic,
    kapture.Keypoints,
    kapture.Descriptors,
    kapture.GlobalFeatures,
    kapture.Matches,
    kapture.Points3d,
    kapture.Observations,
}


def kapture_from_dir(
        kapture_dir_path: str,
        matches_pairs_file_path: Optional[str] = None,
        skip_list: List[Type[Union[kapture.Rigs,
                                   kapture.Trajectories,
                                   kapture.RecordsCamera,
                                   kapture.RecordsDepth,
                                   kapture.RecordsLidar,
                                   kapture.RecordsWifi,
                                   kapture.RecordsBluetooth,
                                   kapture.RecordsGnss,
                                   kapture.RecordsAccelerometer,
                                   kapture.RecordsGyroscope,
                                   kapture.RecordsMagnetic,
                                   kapture.Keypoints,
                                   kapture.Descriptors,
                                   kapture.GlobalFeatures,
                                   kapture.Matches,
                                   kapture.Points3d,
                                   kapture.Observations
                                   ]]] = []
) -> kapture.Kapture:
    """
    Reads and return kapture data from directory.

    :param kapture_dir_path: kapture directory root path
    :param matches_pairs_file_path: text file in the csv format; where each line is image_name1, image_name2, score
    :param skip_list: Input option for expert only. Skip the load of specified parts.
    :return: kapture data read
    """
    if not path.isdir(kapture_dir_path):
        raise FileNotFoundError(f'No kapture directory {kapture_dir_path}')
    csv_file_paths = {dtype: path.join(kapture_dir_path, filename)
                      for dtype, filename in CSV_FILENAMES.items()}
    data_dir_paths = {dtype: path.join(kapture_dir_path, dir_name)
                      for dtype, dir_name in kapture.io.features.FEATURES_DATA_DIRNAMES.items()}

    # keep only those in load_only and that exists
    kapture_data_paths = {**data_dir_paths, **csv_file_paths}  # make sure files take precedence over dirs
    kapture_loadable_data = {
        kapture_type
        for kapture_type in KAPTURE_LOADABLE_TYPES
        if kapture_type not in skip_list and path.exists(kapture_data_paths[kapture_type])
    }

    kapture_data = kapture.Kapture()
    # sensors
    sensor_ids = None
    sensors_file_path = csv_file_paths[kapture.Sensors]
    if sensors_file_path:
        logger.debug(f'loading sensors {sensors_file_path} ...')
        kapture_data.__version__ = get_version_from_csv_file(sensors_file_path)
        kapture_data.sensors = sensors_from_file(sensors_file_path)
        sensor_ids = set(kapture_data.sensors.keys()) if kapture_data.sensors is not None else set()

    if sensor_ids is None:
        # no need to continue, everything else depends on sensors
        raise FileNotFoundError(f'File {sensors_file_path} is missing or empty in {kapture_dir_path}')

    # rigs
    if kapture.Rigs in kapture_loadable_data:
        rigs_file_path = csv_file_paths[kapture.Rigs]
        logger.debug(f'loading rigs {rigs_file_path} ...')
        assert sensor_ids is not None
        kapture_data.rigs = rigs_from_file(rigs_file_path, sensor_ids)
        # update sensor_ids with rig_id
        sensor_ids.update(kapture_data.rigs.keys())

    # trajectories
    if kapture.Trajectories in kapture_loadable_data:
        trajectories_file_path = csv_file_paths[kapture.Trajectories]
        logger.debug(f'loading trajectories {trajectories_file_path} ...')
        assert sensor_ids is not None
        kapture_data.trajectories = trajectories_from_file(trajectories_file_path, sensor_ids)

    _load_all_records(csv_file_paths, kapture_loadable_data, kapture_data)
    _load_features_and_desc_and_matches(data_dir_paths, kapture_dir_path, matches_pairs_file_path,
                                        kapture_loadable_data, kapture_data)
    _load_points3d_and_observations(csv_file_paths, kapture_loadable_data, kapture_data)

    return kapture_data


def get_sensor_ids_of_type(sensor_type: str, sensors: kapture.Sensors) -> Set[str]:
    return set([sensor_id
                for sensor_id in sensors.keys()
                if sensors[sensor_id].sensor_type == sensor_type])


def _load_all_records(csv_file_paths, kapture_loadable_data, kapture_data) -> None:
    """
    Loads all records from disk to the kapture in memory

    :param csv_file_paths: file paths of the CVS records files
    :param kapture_loadable_data: the records to load
    :param kapture_data: to the kapture object to load into
    """
    # records camera
    if kapture.RecordsCamera in kapture_loadable_data:
        records_camera_file_path = csv_file_paths[kapture.RecordsCamera]
        logger.debug(f'loading images {records_camera_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('camera', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_camera = records_camera_from_file(csv_file_paths[kapture.RecordsCamera], sensor_ids)
    # records depth
    if kapture.RecordsDepth in kapture_loadable_data:
        records_depth_file_path = csv_file_paths[kapture.RecordsDepth]
        logger.debug(f'loading depth {records_depth_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('depth', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_depth = records_depth_from_file(csv_file_paths[kapture.RecordsDepth], sensor_ids)
    # records lidar
    if kapture.RecordsLidar in kapture_loadable_data:
        records_lidar_file_path = csv_file_paths[kapture.RecordsLidar]
        logger.debug(f'loading lidar {records_lidar_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('lidar', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_lidar = records_lidar_from_file(records_lidar_file_path, sensor_ids)
    # records Wifi
    if kapture.RecordsWifi in kapture_loadable_data:
        records_wifi_file_path = csv_file_paths[kapture.RecordsWifi]
        logger.debug(f'loading wifi {records_wifi_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('wifi', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_wifi = records_wifi_from_file(records_wifi_file_path, sensor_ids)

    # records bluetooth
    if kapture.RecordsBluetooth in kapture_loadable_data:
        records_bluetooth_file_path = csv_file_paths[kapture.RecordsBluetooth]
        logger.debug(f'loading bluetooth {records_bluetooth_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('bluetooth', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_bluetooth = records_bluetooth_from_file(records_bluetooth_file_path, sensor_ids)

    # records GNSS
    if kapture.RecordsGnss in kapture_loadable_data:
        records_gnss_file_path = csv_file_paths[kapture.RecordsGnss]
        logger.debug(f'loading GNSS {records_gnss_file_path} ...')
        assert kapture_data.sensors is not None
        epsg_codes = {sensor_id: sensor.sensor_params[0]
                      for sensor_id, sensor in kapture_data.sensors.items()
                      if sensor.sensor_type == 'gnss'}
        if len(epsg_codes) > 0:
            kapture_data.records_gnss = records_gnss_from_file(records_gnss_file_path, epsg_codes)
        else:
            logger.warning('no declared GNSS sensors: all GNSS data will be ignored')

    # records Accelerometer
    if kapture.RecordsAccelerometer in kapture_loadable_data:
        records_accelerometer_file_path = csv_file_paths[kapture.RecordsAccelerometer]
        logger.debug(f'loading Accelerations {records_accelerometer_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('accelerometer', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_accelerometer = records_accelerometer_from_file(records_accelerometer_file_path,
                                                                             sensor_ids)
    # records Gyroscope
    if kapture.RecordsGyroscope in kapture_loadable_data:
        records_gyroscope_file_path = csv_file_paths[kapture.RecordsGyroscope]
        logger.debug(f'loading Gyroscope {records_gyroscope_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('gyroscope', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_gyroscope = records_gyroscope_from_file(records_gyroscope_file_path, sensor_ids)
    # records Magnetic
    if kapture.RecordsMagnetic in kapture_loadable_data:
        records_magnetic_file_path = csv_file_paths[kapture.RecordsMagnetic]
        logger.debug(f'loading Magnetic {records_magnetic_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('magnetic', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_magnetic = records_magnetic_from_file(records_magnetic_file_path, sensor_ids)


def _load_features_and_desc_and_matches(data_dir_paths, kapture_dir_path, matches_pairs_file_path,
                                        kapture_loadable_data, kapture_data) -> None:
    """
    Loads features, descriptors, key points and matches from disk to the kapture in memory

    :param data_dir_paths: file paths of the data records files
    :param kapture_dir_path: kapture top directory path
    :param matches_pairs_file_path: text file in the csv format; where each line is image_name1, image_name2, score
    :param kapture_loadable_data: the data to load
    :param kapture_data: to the kapture object to load into
    """

    # features
    image_filenames = set(image_name
                          for _, _, image_name in
                          kapture.flatten(kapture_data.records_camera)) \
        if kapture_data.records_camera is not None else set()
    # keypoints
    if kapture.Keypoints in kapture_loadable_data:
        logger.debug(f'loading keypoints {data_dir_paths[kapture.Keypoints]} ...')
        assert kapture_data.records_camera is not None
        kapture_data.keypoints = keypoints_from_dir(kapture_dir_path, image_filenames)
    # descriptors
    if kapture.Descriptors in kapture_loadable_data:
        logger.debug(f'loading descriptors {data_dir_paths[kapture.Descriptors]} ...')
        assert kapture_data.records_camera is not None
        kapture_data.descriptors = descriptors_from_dir(kapture_dir_path, image_filenames)
    # global_features
    if kapture.GlobalFeatures in kapture_loadable_data:
        logger.debug(f'loading global features {data_dir_paths[kapture.GlobalFeatures]} ...')
        assert kapture_data.records_camera is not None
        kapture_data.global_features = global_features_from_dir(kapture_dir_path, image_filenames)
    # matches
    if kapture.Matches in kapture_loadable_data:
        logger.debug(f'loading matches {data_dir_paths[kapture.Matches]} ...')
        assert kapture_data.records_camera is not None
        kapture_data.matches = matches_from_dir(kapture_dir_path, image_filenames, matches_pairs_file_path)


def _load_points3d_and_observations(csv_file_paths, kapture_loadable_data, kapture_data) -> None:
    # points3d
    if kapture.Points3d in kapture_loadable_data:
        points3d_file_path = csv_file_paths[kapture.Points3d]
        logger.debug(f'loading points 3d {points3d_file_path} ...')
        kapture_data.points3d = points3d_from_file(points3d_file_path)
    # observations
    if kapture.Observations in kapture_loadable_data:
        observations_file_path = csv_file_paths[kapture.Observations]
        logger.debug(f'loading observations {observations_file_path} ...')
        assert kapture_data.keypoints is not None
        assert kapture_data.points3d is not None
        kapture_data.observations = observations_from_file(observations_file_path, kapture_data.keypoints)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from .PoseTransform import PoseTransform
from .Rigs import Rigs
from .flatten import flatten
from typing import Union, Dict, List, Tuple, Optional
from copy import deepcopy


class Trajectories(Dict[int, Dict[str, PoseTransform]]):
    """
    brief: Trajectories
            trajectories[timestamp][sensor_id] = <PoseTransform>
            or
            trajectories[(timestamp, sensor_id)] = <PoseTransform>

            with <PoseTransform> transforming points from world to device.
    """

    def __setitem__(self,
                    key: Union[int, Tuple[int, str]],
                    value: Union[Dict[str, PoseTransform], PoseTransform]):
        # enforce type checking
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            if not isinstance(value, PoseTransform):
                raise TypeError('invalid pose')
            self.setdefault(timestamp, {})[device_id] = value
        elif isinstance(key, int):
            # key is a timestamp
            timestamp = key
            # check type of values?
            if not isinstance(value, dict):
                raise TypeError('invalid value for trajectory timestamp')
            if not all(isinstance(k, str) for k in value.keys()):
                raise TypeError('invalid device_id')
            if not all(isinstance(v, PoseTransform) for v in value.values()):
                raise TypeError('invalid Pose')
            super(Trajectories, self).__setitem__(timestamp, value)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __getitem__(self, key: Union[int, Tuple[int, str]]) -> Union[Dict[str, PoseTransform], PoseTransform]:
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            return super(Trajectories, self).__getitem__(timestamp)[device_id]
        elif isinstance(key, int):
            # key is a timestamp
            return super(Trajectories, self).__getitem__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def key_pairs(self) -> List[Tuple[int, str]]:
        """
        Returns the list of (timestamp, device_id) contained in trajectories.
        Those pairs can be used to access a single trajectory pose.
        :return: list of (timestamp, device_id)
        """
        return [
            (timestamp, sensor_id)
            for timestamp, sensors in self.items()
            for sensor_id in sensors.keys()
        ]

    def __contains__(self, key: Union[int, Tuple[int, str]]):
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            return super(Trajectories, self).__contains__(timestamp) and device_id in self[timestamp]
        elif isinstance(key, int):
            return super(Trajectories, self).__contains__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __repr__(self) -> str:
        # [timestamp, sensor_id] = qw, qx, qy, qz, tx, ty, tz
        lines = [f'[ {timestamp:010}, {sensor_id:5}] = {pose}'
                 for timestamp, sensors in self.items()
                 for sensor_id, pose in sensors.items()]
        return '\n'.join(lines)


def rigs_remove(trajectories: Trajectories, rigs: Rigs) -> Trajectories:
    """
    Returns a new trajectories based on the given one,
    but where rigs poses are replaced by the poses of every sensors.
    This is useful for formats that does not have the rig notion.

    :param trajectories: input Trajectories where the rigs has to be replaced
    :param rigs: input Rigs that defines the rigs/sensors relationship.
    :return: output Trajectories where the rigs has been replaced.
    """
    assert isinstance(rigs, Rigs)
    assert isinstance(trajectories, Trajectories)
    new_trajectories = deepcopy(trajectories)
    rigs_remove_inplace(new_trajectories, rigs)
    return new_trajectories


def rigs_remove_inplace(trajectories: Trajectories, rigs: Rigs, max_depth: int = 10):
    """
    Removes rigs poses and replaces them by the poses of every sensors in it.
    The operation is performed inplace, and modifies trajectories.
    This is useful for formats that does not have the rig notion.

    :param trajectories: input/output Trajectories where the rigs has to be replaced
    :param rigs: input Rigs that defines the rigs/sensors relationship.
    :param max_depth: maximum nested rig depth.
    :return:
    """
    assert isinstance(rigs, Rigs)
    assert isinstance(trajectories, Trajectories)
    # collect all poses of rigs in trajectories
    for iteration in range(max_depth):
        # repeat the operation while there is so rig remaining (nested rigs)
        jobs = [(timestamp, rig_id, pose_rig_from_world)
                for timestamp, rig_id, pose_rig_from_world in flatten(trajectories)
                if rig_id in rigs.keys()]

        if len(jobs) == 0:
            # we are done
            break

        # replace those rigs poses by the one of the sensors
        for timestamp, rig_id, pose_rig_from_world in jobs:
            # its a rig, add every sensors in it instead.
            for device_id, pose_device_from_rig in rigs[rig_id].items():
                pose_cam_from_world = PoseTransform.compose([pose_device_from_rig, pose_rig_from_world])
                trajectories[timestamp, device_id] = pose_cam_from_world
            # then remove this rig pose
            del trajectories[timestamp][rig_id]

    # remove useless (empty) timestamp (if any) from trajectories
    for timestamp in trajectories.keys():
        if len(trajectories[timestamp]) == 0:
            del trajectories[timestamp]

    # Do not clear, the rigs : so easy to do outside, and so easy


def rigs_recover(
        trajectories: Trajectories,
        rigs: Rigs,
        master_sensors: Optional[List[str]] = None
) -> Trajectories:
    """
    Returns a new trajectories based on the given one,
    but where sensor poses are replaced by the rig pose, each time the sensor is registered in a rig.
    The rig configuration must be given.
    Warning: in some configuration, it may leads to inconsistent results. See code for details.

    :param trajectories: input Trajectories.
    :param rigs: input Rigs configuration.
    :param master_sensors: input If given, only compute rig poses for the given sensors.
    :return: a new Trajectories.
    """
    new_trajectories = deepcopy(trajectories)
    rigs_recover_inplace(new_trajectories, rigs, master_sensors)
    return new_trajectories


def rigs_recover_inplace(
        trajectories: Trajectories,
        rigs: Rigs,
        master_sensors: Optional[List[str]] = None,
        max_depth: int = 10
):
    """
    Updates the given trajectories by replacing sensor poses by the rig poses,
    each time the sensor is registered in a rig.
    The rig configuration must be given.
    Warning: in some configuration, it may leads to inconsistent results. See code for details.

    :param trajectories: input/output Trajectories.
    :param rigs: input Rigs configuration.
    :param master_sensors: input If given, only compute rig poses for the given sensors.
    :return:
    """

    # sensor_id -> rig_id, pose_rig_from_sensor
    reverse_rig_dict = {
        sensor_id: (rig_id, rigs[rig_id, sensor_id].inverse())
        for rig_id, sensor_id in rigs.key_pairs()
    }

    for iteration in range(max_depth):
        # do the replacement while there is sensor_id in trajectories, that can be converted to rig_id
        jobs = [(timestamp, sensor_id, pose_sensor_from_world)
                for timestamp, sensor_id, pose_sensor_from_world in flatten(trajectories, is_sorted=True)
                if sensor_id in reverse_rig_dict]

        if len(jobs) == 0:
            break

        for timestamp, sensor_id, pose_sensor_from_world in jobs:
            # if the sensor is part of a rig, set the pose of the rig,
            # instead of the pose of the sensor
            rig_id, pose_rig_from_sensor = reverse_rig_dict[sensor_id]
            pose_sensor_from_world = trajectories[timestamp].pop(sensor_id)

            # only compute rig poses for master_sensors if any.
            if master_sensors is not None and sensor_id not in master_sensors:
                continue

            # skip if rig pose already recovered.
            if rig_id in trajectories[timestamp]:
                continue

            # warning: if multiple sensors can be used to infer the rig pose
            # (eg. no master sensor, or multiple master sensors), only one sensor is used to compute the rig pose.
            # Since we use is_sorted=True, the selected sensor should be consistent, when present. But if for some
            # timestamp, the usual first sensor is missing, then, another sensor is used as reference and
            # if the sensors dos not actually use the rig calibration (no rigid transform between sensors),
            # it may end up to inconsistent results.
            pose_rig_from_world = PoseTransform.compose([pose_rig_from_sensor, pose_sensor_from_world])
            trajectories[timestamp, rig_id] = pose_rig_from_world



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Merge kapture objects for reconstructions.
"""

import numpy as np
import os
import shutil
from typing import List, Union, Optional, Tuple, Type

import kapture
import kapture.io.features
from kapture.utils.logging import getLogger


def merge_image_features(feature_type: Type[Union[kapture.Keypoints, kapture.Descriptors, kapture.GlobalFeatures]],
                         features_list: Union[List[Optional[kapture.Keypoints]],
                                              List[Optional[kapture.Descriptors]],
                                              List[Optional[kapture.GlobalFeatures]]],
                         features_paths: List[str],
                         output_path: str
                         ) -> Union[kapture.Keypoints, kapture.Descriptors, kapture.GlobalFeatures]:
    """
    Merge several features_list (keypoints, descriptors or global features_list) (of same type) in one.

    :param feature_type: the type of features_list
    :param features_list: the list of values
    :param features_paths: the paths
    :param output_path: root path of the features to construct
    :return: merged features object of the corresponding type
    """
    assert len(features_list) > 0
    assert len(features_paths) == len(features_list)

    # find no none value
    val = [d for d in features_list if d is not None]
    assert len(val) > 0

    merged_features = feature_type(val[0].type_name, val[0].dtype, val[0].dsize)
    for features, features_path in zip(features_list, features_paths):
        if features is None:
            continue
        assert isinstance(features, feature_type)
        assert features.type_name == merged_features.type_name
        assert features.dtype == merged_features.dtype
        assert features.dsize == merged_features.dsize
        for name in features:
            if name in merged_features:
                getLogger().warning(f'{name} was found multiple times.')
            else:
                merged_features.add(name)
                if output_path:
                    # TODO: uses kapture.io.features_list.get_image_features_dirpath()
                    in_path = kapture.io.features.get_features_fullpath(feature_type, features_path, name)
                    out_path = kapture.io.features.get_features_fullpath(feature_type, output_path, name)
                    if in_path != out_path:
                        # skip actual copy if file does not actually move.
                        os.makedirs(os.path.dirname(out_path), exist_ok=True)
                        shutil.copy(in_path, out_path)
    return merged_features


def merge_keypoints(keypoints_list: List[Optional[kapture.Keypoints]],
                    keypoints_paths: List[str],
                    output_path: str) -> kapture.Keypoints:
    """
    Merge several keypoints in one.

    :param keypoints_list: list of keypoints to merge
    :param keypoints_paths: keypoints files paths
    :param output_path: root path of the merged features files
    :return: merged keypoints
    """
    keypoints = merge_image_features(kapture.Keypoints, keypoints_list, keypoints_paths, output_path)
    assert isinstance(keypoints, kapture.Keypoints)
    return keypoints


def merge_descriptors(descriptors_list: List[Optional[kapture.Descriptors]],
                      descriptors_paths: List[str], output_path: str) -> kapture.Descriptors:
    """
    Merge several descriptors in one.

    :param descriptors_list: list of descriptors to merge
    :param descriptors_paths: descriptors files paths
    :param output_path: root path of the merged features files
    :return: merged descriptors
    """
    descriptors = merge_image_features(kapture.Descriptors, descriptors_list, descriptors_paths, output_path)
    assert isinstance(descriptors, kapture.Descriptors)
    return descriptors


def merge_global_features(global_features_list: List[Optional[kapture.GlobalFeatures]],
                          global_features_paths: List[str], output_path: str) -> kapture.GlobalFeatures:
    """
    Merge several global features in one.

    :param global_features_list: list of global features to merge
    :param global_features_paths: global features files paths
    :param output_path: root path of the merged features files
    :return: merged global features
    """
    features = merge_image_features(kapture.GlobalFeatures, global_features_list, global_features_paths, output_path)
    assert isinstance(features, kapture.GlobalFeatures)
    return features


def merge_matches(matches_list: List[Optional[kapture.Matches]],
                  matches_paths: List[str],
                  output_path: str) -> kapture.Matches:
    """
    Merge several matches lists in one.

    :param matches_list: list of matches to merge
    :param matches_paths: matches files paths
    :param output_path: root path of the merged matches files
    :return: merged matches
    """
    assert len(matches_list) > 0
    assert len(matches_paths) == len(matches_list)

    merged_matches = kapture.Matches()
    for matches, matches_path in zip(matches_list, matches_paths):
        if matches is None:
            continue
        for pair in matches:
            if pair in merged_matches:
                getLogger().warning(f'{pair} was found multiple times.')
            else:
                merged_matches.add(pair[0], pair[1])
                if output_path:
                    in_path = kapture.io.features.get_matches_fullpath(pair, matches_path)
                    out_path = kapture.io.features.get_matches_fullpath(pair, output_path)
                    if in_path != out_path:
                        # skip actual copy if file does not actually move.
                        os.makedirs(os.path.dirname(out_path), exist_ok=True)
                        shutil.copy(in_path, out_path)
    return merged_matches


def merge_points3d_and_observations(pts3d_obs: List[Tuple[Optional[kapture.Points3d], Optional[kapture.Observations]]]
                                    ) -> Tuple[kapture.Points3d, kapture.Observations]:
    """
    Merge a list of points3d with their observations.

    :param pts3d_obs: list of points3d with observations to merge
    :return: merged points3d associated to observations
    """
    assert len(pts3d_obs) > 0
    merged_points3d = kapture.Points3d()
    merged_observations = kapture.Observations()
    point3d_offset = 0
    for points3d, observations in pts3d_obs:
        if points3d is None:
            continue
        merged_points3d = kapture.Points3d(np.vstack([merged_points3d, points3d]))
        if observations is not None:
            for point3d_idx, (image_path, keypoint_idx) in kapture.flatten(observations):
                merged_observations.add(point3d_idx + point3d_offset, image_path, keypoint_idx)
        point3d_offset += merged_points3d.shape[0]
    return merged_points3d, merged_observations


def merge_points3d(points3d_list: List[Optional[kapture.Points3d]]) -> kapture.Points3d:
    """
    Merge several points3d lists in one.

    :param points3d_list: list of points3d to merge
    :return: merged points3d
    """
    assert len(points3d_list) > 0
    merged_points3d = kapture.Points3d()
    for points3d in points3d_list:
        if points3d is None:
            continue
        merged_points3d = kapture.Points3d(np.vstack([merged_points3d, points3d]))
    return merged_points3d



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Kapture to openmvg export functions.
"""

import json
import logging
import os
import os.path as path
from typing import Dict, List, Optional, Union
import quaternion
import numpy as np
from tqdm import tqdm
# kapture
import kapture
import kapture.io.csv
from kapture.io.binary import TransferAction, transfer_files_from_dir, array_to_file
from kapture.io.records import get_image_fullpath
from kapture.io.features import keypoints_to_filepaths, image_keypoints_from_file, get_keypoints_fullpath
from kapture.io.features import descriptors_to_filepaths, image_descriptors_from_file
from kapture.io.features import matches_to_filepaths, image_matches_from_file
import kapture.io.structure
from kapture.core.Trajectories import rigs_remove_inplace
from kapture.utils.paths import safe_remove_file, safe_remove_any_path
# local
from .openmvg_commons import JSON_KEY, OPENMVG_SFM_DATA_VERSION_NUMBER, OPENMVG_DEFAULT_JSON_FILE_NAME
from .openmvg_commons import CameraModel

logger = logging.getLogger('openmvg')  # Using global openmvg logger

NEW_ID_MASK = 1 << 31  # 10000000 00000000 00000000 00000000
DEFAULT_FOCAL_LENGTH_FACTOR = 1.2


class CerealPointerRegistry:
    NEW_ID_MASK = 1 << 31  # 10000000 00000000 00000000 00000000
    NULL_ID = 1 << 30  # 01000000 00000000 00000000 00000000

    def __init__(self, value_key, id_key):
        self._value_key = value_key
        self._id_key = id_key
        self._ids = {}
        self._id_current = 1

    def get_ids_dict(self, value):
        """ :return either id if known or id + name if new """
        if isinstance(value, dict) or value not in self._ids:
            # if this is the first time or a dict
            new_id = self._id_current
            if not isinstance(value, dict):
                self._ids[value] = new_id
            self._id_current += 1
            return {
                self._id_key: new_id | NEW_ID_MASK,
                self._value_key: value,
            }
        else:
            return {
                self._id_key: self._ids[value]
            }


def _get_data(camera_params: list) -> Dict:
    # w, h, f, cx, cy
    data: Dict[str, Union[int, float, List[float]]] = {JSON_KEY.WIDTH: int(camera_params[0]),
                                                       JSON_KEY.HEIGHT: int(camera_params[1]),
                                                       JSON_KEY.FOCAL_LENGTH: float(camera_params[2]),
                                                       JSON_KEY.PRINCIPAL_POINT: [float(camera_params[3]),
                                                                                  float(camera_params[4])]}
    return data


def _get_intrinsic_pinhole(camera_params: list) -> Dict:
    # w, h, f, cx, cy
    return _get_data(camera_params)


def _get_intrinsic_pinhole_radial_k1(camera_params: list) -> Dict:
    data = _get_data(camera_params)
    # w, h, f, cx, cy, k
    data[JSON_KEY.DISTO_K1] = [float(camera_params[5])]
    return data


def _get_intrinsic_pinhole_radial_k3(camera_params: list) -> Dict:
    data = _get_data(camera_params)
    # w, h, f, cx, cy, k1, k2, k3
    data[JSON_KEY.DISTO_K3] = [float(camera_params[5]), float(camera_params[6]), float(camera_params[7])]
    return data


def _get_intrinsic_pinhole_brown_t2(camera_params: list) -> Dict:
    # w, h, f, cx, cy, k1, k2, k3, t1, t2
    disto_t2 = [float(camera_params[5]), float(camera_params[6]), float(camera_params[7]),
                float(camera_params[8]), float(camera_params[9])]
    return {JSON_KEY.VALUE0: _get_data(camera_params), JSON_KEY.DISTO_T2: disto_t2}


def _get_intrinsic_fisheye(camera_params: list) -> Dict:
    # w, h, f, cx, cy, k1, k2, k3, k4
    fisheye = [float(camera_params[5]), float(camera_params[6]), float(camera_params[7]), float(camera_params[8])]
    return {JSON_KEY.VALUE0: _get_data(camera_params),
            JSON_KEY.FISHEYE: fisheye}


def get_openmvg_camera_id(kapture_camera_id, kapture_to_openmvg_cam_ids):
    """ return a valid openmvg camera id for the given kapture on.
    It keeps kapture_to_openmvg_cam_ids uptodate, and ensure there is no collision.
    """

    if kapture_camera_id in kapture_to_openmvg_cam_ids:
        # already defined
        return kapture_to_openmvg_cam_ids[kapture_camera_id]

    # its not defined, then, make up one
    last_known_openmvg_camera_id = max(list(kapture_to_openmvg_cam_ids.values()) + [-1])
    assert type(last_known_openmvg_camera_id) == int
    openmvg_camera_id = last_known_openmvg_camera_id + 1
    assert openmvg_camera_id not in kapture_to_openmvg_cam_ids
    kapture_to_openmvg_cam_ids[kapture_camera_id] = openmvg_camera_id
    return openmvg_camera_id


def get_openmvg_image_path(kapture_image_name: str, flatten_path: bool = False):
    """ the openmvg image sub path corresponding to the given kapture one. """
    return kapture_image_name if not flatten_path else kapture_image_name.replace('/', '_')


def export_openmvg_intrinsics(
        kapture_cameras,
        kapture_to_openmvg_cam_ids: Dict[str, int],
        polymorphic_registry: CerealPointerRegistry,
        ptr_wrapper_registry: CerealPointerRegistry,
):
    """
    Exports the given kapture cameras to the openMVG sfm_data structure.
    In openMVGm, cameras are referred as Intrinsics camera internal parameters.

    :param kapture_cameras: input kapture cameras to be exported (even if not used).
    :param kapture_to_openmvg_cam_ids: input/output dict that maps kapture camera ids to openMVG camera ids.
    :param polymorphic_registry: input/output polymorphic IDs status
    :param ptr_wrapper_registry: input/output polymorphic IDs status
    :return:
    """
    openmvg_intrinsics = []
    # kapture_to_openmvg_cam_ids = {}
    # process all cameras
    for kapture_cam_id, kapture_camera in kapture_cameras.items():
        openmvg_camera_id = get_openmvg_camera_id(kapture_cam_id, kapture_to_openmvg_cam_ids)
        kapture_cam_type = kapture_camera.camera_type
        kapture_camera_params = kapture_camera.camera_params
        if kapture_cam_type == kapture.CameraType.SIMPLE_PINHOLE:
            # w, h, f, cx, cy
            opnmvg_cam_type = CameraModel.pinhole
            data = _get_intrinsic_pinhole(kapture_camera_params)
        elif kapture_cam_type == kapture.CameraType.PINHOLE:
            # w, h, f, cx, cy
            opnmvg_cam_type = CameraModel.pinhole
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            (kapture_camera_params[2] + kapture_camera_params[3]) / 2,  # fx+fy/2 as f
                            kapture_camera_params[4], kapture_camera_params[5]]  # cx cy
            data = _get_intrinsic_pinhole(faked_params)
        elif kapture_cam_type == kapture.CameraType.SIMPLE_RADIAL:
            # w, h, f, cx, cy, k
            opnmvg_cam_type = CameraModel.pinhole_radial_k1
            data = _get_intrinsic_pinhole_radial_k1(kapture_camera_params)
        elif kapture_cam_type == kapture.CameraType.RADIAL:
            # w, h, f, cx, cy, k1, k2, k3
            opnmvg_cam_type = CameraModel.pinhole_radial_k3
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            kapture_camera_params[2],  # f
                            kapture_camera_params[3], kapture_camera_params[4],  # cx cy
                            kapture_camera_params[5], kapture_camera_params[6], 0  # k1, k2, k3
                            ]
            data = _get_intrinsic_pinhole_radial_k3(faked_params)
        elif kapture_cam_type == kapture.CameraType.FULL_OPENCV or kapture_cam_type == kapture.CameraType.OPENCV:
            # w, h, f, cx, cy, k1, k2, k3, t1, t2
            opnmvg_cam_type = CameraModel.pinhole_brown_t2
            k3 = kapture_camera_params[10] if len(kapture_camera_params) > 10 else 0
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            (kapture_camera_params[2] + kapture_camera_params[3]) / 2,  # fx+fy/2 as f
                            kapture_camera_params[4], kapture_camera_params[5],  # cx cy
                            kapture_camera_params[6], kapture_camera_params[7], k3,  # k1, k2, k3
                            kapture_camera_params[8], kapture_camera_params[9]  # p1, p2 (=t1, t2)
                            ]
            data = _get_intrinsic_pinhole_brown_t2(faked_params)
        elif kapture_cam_type == kapture.CameraType.OPENCV_FISHEYE:
            logger.warning('OpenCV fisheye model is not compatible with OpenMVG. Forcing distortion to 0')
            # w, h, f, cx, cy, k1, k2, k3, k4
            opnmvg_cam_type = CameraModel.fisheye
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            (kapture_camera_params[2] + kapture_camera_params[3]) / 2,  # fx+fy/2 as f
                            kapture_camera_params[4], kapture_camera_params[5],  # cx cy
                            0, 0,  # k1, k2
                            0, 0  # k3, k4
                            ]
            data = _get_intrinsic_fisheye(faked_params)
        elif kapture_cam_type == kapture.CameraType.RADIAL_FISHEYE or \
                kapture_cam_type == kapture.CameraType.SIMPLE_RADIAL_FISHEYE:
            logger.warning('OpenCV fisheye model is not compatible with OpenMVG. Forcing distortion to 0')
            # w, h, f, cx, cy, k1, k2, k3, k4
            opnmvg_cam_type = CameraModel.fisheye
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            kapture_camera_params[2],  # f
                            kapture_camera_params[3], kapture_camera_params[4],  # cx cy
                            0, 0,  # k1, k2
                            0, 0  # k3, k4
                            ]
            data = _get_intrinsic_fisheye(faked_params)
        elif kapture_cam_type == kapture.CameraType.UNKNOWN_CAMERA:
            logger.info(f'Camera {kapture_cam_id}: Unknown camera model, using simple radial')
            # Choose simple radial model, to allow openMVG to determine distortion param
            # w, h, f, cx, cy, k
            opnmvg_cam_type = CameraModel.pinhole_radial_k1
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            max(kapture_camera_params[0], kapture_camera_params[1]) * DEFAULT_FOCAL_LENGTH_FACTOR,
                            # max(w,h)*1.2 as f
                            int(kapture_camera_params[0] / 2), int(kapture_camera_params[1] / 2),  # cx cy
                            0.0]  # k1
            data = _get_intrinsic_pinhole_radial_k1(faked_params)
        else:
            raise ValueError(f'Camera model {kapture_cam_type.value} not supported')

        intrinsic = {}
        cam_type_poly_id = polymorphic_registry.get_ids_dict(opnmvg_cam_type.name)
        intrinsic.update(cam_type_poly_id)

        # it is assumed that this camera is only encountered once
        # set the first bit of ptr_wrapper_id_current to 1
        data_wrapper = ptr_wrapper_registry.get_ids_dict(data)

        intrinsic[JSON_KEY.PTR_WRAPPER] = data_wrapper
        openmvg_intrinsics.append({JSON_KEY.KEY: openmvg_camera_id, JSON_KEY.VALUE: intrinsic})

    return openmvg_intrinsics


def export_openmvg_views(
        kapture_cameras: kapture.Sensors,
        kapture_images: kapture.RecordsCamera,
        kapture_trajectories: kapture.Trajectories,
        kapture_to_openmvg_cam_ids: Dict[str, int],
        kapture_to_openmvg_view_ids: Dict[str, int],
        polymorphic_registry: CerealPointerRegistry,
        ptr_wrapper_registry: CerealPointerRegistry,
        image_path_flatten: bool,
):
    """

    :param kapture_cameras:
    :param kapture_images:
    :param kapture_trajectories:
    :param kapture_to_openmvg_cam_ids: input dict that maps kapture camera ids to openMVG camera ids.
    :param kapture_to_openmvg_view_ids: input dict that maps kapture image names to openMVG view ids.
    :param polymorphic_registry: input/output polymorphic IDs status
    :param ptr_wrapper_registry: input/output polymorphic IDs status
    :param image_path_flatten: flatten image path (eg. to avoid image name collision in openMVG regions).
    :return:
    """
    views = []
    # process all images
    for timestamp, kapture_cam_id, kapture_image_name in kapture.flatten(kapture_images):
        assert kapture_cam_id in kapture_to_openmvg_cam_ids
        assert kapture_image_name in kapture_to_openmvg_view_ids
        openmvg_cam_id = get_openmvg_camera_id(kapture_cam_id, kapture_to_openmvg_cam_ids)
        openmvg_view_id = kapture_to_openmvg_view_ids[kapture_image_name]
        openmvg_image_filepath = get_openmvg_image_path(kapture_image_name, image_path_flatten)
        openmvg_image_filename = path.basename(openmvg_image_filepath)
        openmvg_image_local_path = path.dirname(openmvg_image_filepath)
        kapture_camera_params = kapture_cameras[kapture_cam_id].camera_params
        view_data = {JSON_KEY.LOCAL_PATH: openmvg_image_local_path,
                     JSON_KEY.FILENAME: openmvg_image_filename,
                     JSON_KEY.WIDTH: int(kapture_camera_params[0]),
                     JSON_KEY.HEIGHT: int(kapture_camera_params[1]),
                     JSON_KEY.ID_VIEW: openmvg_view_id,
                     JSON_KEY.ID_INTRINSIC: openmvg_cam_id,
                     JSON_KEY.ID_POSE: openmvg_view_id}

        view = {}
        # retrieve image pose from trajectories
        if timestamp not in kapture_trajectories:
            view[JSON_KEY.POLYMORPHIC_ID] = CerealPointerRegistry.NULL_ID
        else:
            # there is a pose for that timestamp
            # The poses are stored both as priors (in the 'views' table) and as known poses (in the 'extrinsics' table)
            assert kapture_cam_id in kapture_trajectories[timestamp]
            view_priors_id = polymorphic_registry.get_ids_dict(JSON_KEY.VIEW_PRIORS)
            view.update(view_priors_id)

            pose_tr = kapture_trajectories[timestamp].get(kapture_cam_id)
            prior_q = pose_tr.r
            prior_t = pose_tr.inverse().t_raw
            pose_data = {JSON_KEY.CENTER: prior_t,
                         JSON_KEY.ROTATION: quaternion.as_rotation_matrix(prior_q).tolist()}

            view_data[JSON_KEY.USE_POSE_CENTER_PRIOR] = True
            view_data[JSON_KEY.CENTER_WEIGHT] = [1.0, 1.0, 1.0]
            view_data[JSON_KEY.CENTER] = prior_t
            view_data[JSON_KEY.USE_POSE_ROTATION_PRIOR] = True
            view_data[JSON_KEY.ROTATION_WEIGHT] = 1.0
            view_data[JSON_KEY.ROTATION] = pose_data[JSON_KEY.ROTATION]

        # it is assumed that this view is only encountered once
        view_wrapper = ptr_wrapper_registry.get_ids_dict(view_data)
        view[JSON_KEY.PTR_WRAPPER] = view_wrapper
        views.append({JSON_KEY.KEY: openmvg_view_id, JSON_KEY.VALUE: view})

    return views


def export_openmvg_poses(
        kapture_images: kapture.RecordsCamera,
        kapture_trajectories: kapture.Trajectories,
        kapture_to_openmvg_view_ids: Dict[str, int],
):
    """

    :param kapture_images:
    :param kapture_trajectories:
    :param kapture_to_openmvg_view_ids: input dict that maps kapture image ids to openMVG view ids.
    :return:
    """
    extrinsics = []
    # process all images
    for timestamp, kapture_cam_id, kapture_image_name in kapture.flatten(kapture_images):
        assert kapture_image_name in kapture_to_openmvg_view_ids
        openmvg_view_id = kapture_to_openmvg_view_ids[kapture_image_name]
        # retrieve image pose from trajectories
        if timestamp in kapture_trajectories:
            # there is a pose for that timestamp
            # The poses are stored both as priors (in the 'views' table) and as known poses (in the 'extrinsics' table)
            assert kapture_cam_id in kapture_trajectories[timestamp]
            pose_tr = kapture_trajectories[timestamp].get(kapture_cam_id)
            prior_q = pose_tr.r
            prior_t = pose_tr.inverse().t_raw
            pose_data = {JSON_KEY.CENTER: prior_t,
                         JSON_KEY.ROTATION: quaternion.as_rotation_matrix(prior_q).tolist()}
            extrinsics.append({JSON_KEY.KEY: openmvg_view_id, JSON_KEY.VALUE: pose_data})

    return extrinsics


def export_openmvg_structure(
        kapture_points_3d: kapture.Points3d,
        kapture_to_openmvg_view_ids: Dict[str, int],
        kapture_observations: Optional[kapture.Observations] = None,
        kapture_keypoints: Optional[kapture.Keypoints] = None,
        kapture_path: Optional[str] = None,
):
    # early check
    if kapture_points_3d is None:
        logger.warning(f'no 3D points to export.')
        return

    xyz_coordinates = kapture_points_3d[:, 0:3]
    include_2d_observations = kapture_observations is not None
    openmvg_structure = []
    # this loop can be very long, lets show some progress
    hide_progress_bars = logger.getEffectiveLevel() > logging.INFO

    for point_idx, coords in enumerate(tqdm(xyz_coordinates, disable=hide_progress_bars)):
        point_3d_structure = {
            'key': point_idx,
            'value': {
                'X': coords.tolist(),
                'observations': []
            }
        }
        if include_2d_observations and point_idx in kapture_observations:
            for kapture_image_name, feature_point_id in kapture_observations[point_idx]:
                openmvg_view_id = kapture_to_openmvg_view_ids[kapture_image_name]
                point_2d_observation = {'key': openmvg_view_id,
                                        'value': {'id_feat': feature_point_id, }}

                if kapture_path and kapture_keypoints is not None:
                    # if given, load keypoints to populate 2D coordinates of the feature.
                    keypoints_file_path = get_keypoints_fullpath(kapture_path, kapture_image_name)
                    try:
                        keypoints_data = image_keypoints_from_file(keypoints_file_path,
                                                                   dtype=kapture_keypoints.dtype,
                                                                   dsize=kapture_keypoints.dsize)
                        point_2d_observation['value']['x'] = keypoints_data[feature_point_id, 0:2].tolist()
                    except FileNotFoundError:
                        logger.warning(f'unable to load keypoints file {keypoints_file_path}')

                point_3d_structure['value']['observations'].append(point_2d_observation)

        openmvg_structure.append(point_3d_structure)

    return openmvg_structure


def export_openmvg_sfm_data(
        kapture_path: str,
        kapture_data: kapture.Kapture,
        openmvg_sfm_data_file_path: str,
        openmvg_image_root_path: str,
        image_action: TransferAction,
        image_path_flatten: bool,
        force: bool,
        kapture_to_openmvg_view_ids: dict = {}
) -> Dict:
    """
    Convert the kapture data into an openMVG dataset stored as a dictionary.
    The format is defined here:
    https://openmvg.readthedocs.io/en/latest/software/SfM/SfM_OutputFormat/

    :param kapture_data: the kapture data
    :param kapture_path: top directory of the kapture data and the images
    :param openmvg_sfm_data_file_path: input path to the SfM data file to be written.
    :param openmvg_image_root_path: input path to openMVG image directory to be created.
    :param image_action: action to apply on images: link, copy, move or do nothing.
    :param image_path_flatten: flatten image path (eg. to avoid image name collision in openMVG regions).
    :param force: if true, will remove existing openMVG data without prompting the user.
    :param kapture_to_openmvg_view_ids: input/output mapping of kapture image name to corresponding openmvg view id.
    :return: an SfM_data, the openmvg structure, stored as a dictionary ready to be serialized
    """

    if kapture_data.cameras is None or kapture_data.records_camera is None:
        raise ValueError('export_openmvg_sfm_data needs kapture camera and records_camera.')

    if image_action == TransferAction.root_link:
        raise NotImplementedError('root link is not implemented, use skip instead.')

    # refer to the original image dir when skipping image transfer.
    if image_action == TransferAction.skip:
        openmvg_image_root_path = get_image_fullpath(kapture_path)

    if openmvg_image_root_path is None:
        raise ValueError(f'openmvg_image_root_path must be defined to be able to perform {image_action}.')

    # make sure directory is ready to contain openmvg_sfm_data_file_path
    os.makedirs(path.dirname(openmvg_sfm_data_file_path), exist_ok=True)

    # Check we don't have other sensors defined
    if len(kapture_data.sensors) != len(kapture_data.cameras):
        extra_sensor_number = len(kapture_data.sensors) - len(kapture_data.cameras)
        logger.warning(f'We will ignore {extra_sensor_number} sensors that are not camera')

    # openmvg does not support rigs
    if kapture_data.rigs:
        logger.info('remove rigs notation.')
        rigs_remove_inplace(kapture_data.trajectories, kapture_data.rigs)
        kapture_data.rigs.clear()

    # Compute root path and camera used in records
    kapture_to_openmvg_cam_ids = {}  # kapture_cam_id -> openmvg_cam_id
    for i, (_, _, kapture_image_name) in enumerate(kapture.flatten(kapture_data.records_camera)):
        if kapture_image_name not in kapture_to_openmvg_view_ids:
            kapture_to_openmvg_view_ids[kapture_image_name] = i

    # polymorphic_status = PolymorphicStatus({}, 1, 1)
    polymorphic_registry = CerealPointerRegistry(id_key=JSON_KEY.POLYMORPHIC_ID, value_key=JSON_KEY.POLYMORPHIC_NAME)
    ptr_wrapper_registry = CerealPointerRegistry(id_key=JSON_KEY.ID, value_key=JSON_KEY.DATA)

    logger.debug(f'exporting intrinsics ...')
    openmvg_sfm_data_intrinsics = export_openmvg_intrinsics(
        kapture_cameras=kapture_data.cameras,
        kapture_to_openmvg_cam_ids=kapture_to_openmvg_cam_ids,
        polymorphic_registry=polymorphic_registry,
        ptr_wrapper_registry=ptr_wrapper_registry,
    )

    logger.debug(f'exporting views ...')
    openmvg_sfm_data_views = export_openmvg_views(
        kapture_cameras=kapture_data.cameras,
        kapture_images=kapture_data.records_camera,
        kapture_trajectories=kapture_data.trajectories,
        kapture_to_openmvg_cam_ids=kapture_to_openmvg_cam_ids,
        kapture_to_openmvg_view_ids=kapture_to_openmvg_view_ids,
        polymorphic_registry=polymorphic_registry,
        ptr_wrapper_registry=ptr_wrapper_registry,
        image_path_flatten=image_path_flatten,
    )

    logger.debug(f'exporting poses ...')
    openmvg_sfm_data_poses = export_openmvg_poses(
        kapture_images=kapture_data.records_camera,
        kapture_trajectories=kapture_data.trajectories,
        kapture_to_openmvg_view_ids=kapture_to_openmvg_view_ids)

    # structure : correspond to kapture observations + 3D points
    logger.debug(f'exporting structure ...')
    openmvg_sfm_data_structure = export_openmvg_structure(
        kapture_points_3d=kapture_data.points3d,
        kapture_to_openmvg_view_ids=kapture_to_openmvg_view_ids,
        kapture_observations=kapture_data.observations,
        kapture_keypoints=kapture_data.keypoints,
        kapture_path=kapture_path
    )

    openmvg_sfm_data = {
        JSON_KEY.SFM_DATA_VERSION: OPENMVG_SFM_DATA_VERSION_NUMBER,
        JSON_KEY.ROOT_PATH: path.abspath(openmvg_image_root_path),
        JSON_KEY.INTRINSICS: openmvg_sfm_data_intrinsics,
        JSON_KEY.VIEWS: openmvg_sfm_data_views,
        JSON_KEY.EXTRINSICS: openmvg_sfm_data_poses,
        JSON_KEY.STRUCTURE: openmvg_sfm_data_structure,
        JSON_KEY.CONTROL_POINTS: [],
    }

    logger.debug(f'Saving to openmvg {openmvg_sfm_data_file_path}...')
    with open(openmvg_sfm_data_file_path, "w") as fid:
        json.dump(openmvg_sfm_data, fid, indent=4)

    # do the actual image transfer
    if not image_action == TransferAction.skip:
        job_copy = (
            (  # source path -> dest path
                get_image_fullpath(kapture_path, kapture_image_name),
                path.join(openmvg_image_root_path, get_openmvg_image_path(kapture_image_name, image_path_flatten))
            )
            for _, _, kapture_image_name in kapture.flatten(kapture_data.records_camera)
        )
        source_filepath_list, destination_filepath_list = zip(*job_copy)
        transfer_files_from_dir(
            source_filepath_list=source_filepath_list,
            destination_filepath_list=destination_filepath_list,
            copy_strategy=image_action,
            force_overwrite=force
        )


def export_openmvg_regions(
        kapture_path: str,
        kapture_keypoints: Optional[kapture.Keypoints],
        kapture_descriptors: kapture.Descriptors,
        openmvg_regions_dir_path: str,
        image_path_flatten: bool
):
    """
    exports openMVG regions ie keypoints and descriptors.

    :param kapture_path: input path to root kapture directory.
    :param kapture_keypoints: input kapture keypoints. Could be None if no keypoints.
    :param kapture_descriptors: input kapture descriptors. Could be None if no descriptors.
    :param openmvg_regions_dir_path: input path to output openMVG regions directory.
    :param image_path_flatten: if true, it means that image path are to be flatten.
    :return:
    """
    # early check we should do
    if kapture_keypoints is None or kapture_descriptors is None:
        logger.warning('no keypoints or descriptors to export.')
        return

    # make sure output directory is ready
    os.makedirs(openmvg_regions_dir_path, exist_ok=True)

    # only able to export SIFT
    if any([f.type_name.upper() != 'SIFT' for f in [kapture_keypoints, kapture_descriptors]]):
        raise ValueError(f'unable to export other regions than sift '
                         f'(got {kapture_keypoints.type_name}/{kapture_descriptors.type_name})')

    os.makedirs(openmvg_regions_dir_path, exist_ok=True)
    polymorphic_registry = CerealPointerRegistry(id_key=JSON_KEY.POLYMORPHIC_ID, value_key=JSON_KEY.POLYMORPHIC_NAME)
    # create image_describer.json
    fake_regions_type = {"ptr_wrapper": {"valid": 1, "data": {"value0": [], "value1": []}}}
    fake_regions_type.update(polymorphic_registry.get_ids_dict('SIFT_Regions'))
    image_describer = {
        'regions_type': fake_regions_type
    }
    image_describer_file_path = path.join(openmvg_regions_dir_path, 'image_describer.json')
    with open(image_describer_file_path, 'w') as fid:
        json.dump(image_describer, fid, indent=4)

    # this loop can be very long, lets show some progress
    hide_progress_bars = logger.getEffectiveLevel() > logging.INFO

    # copy keypoints files
    keypoints = keypoints_to_filepaths(kapture_keypoints, kapture_path)
    for kapture_image_name, kapture_keypoint_file_path in tqdm(keypoints.items(), disable=hide_progress_bars):
        openmvg_keypoint_file_name = get_openmvg_image_path(kapture_image_name, image_path_flatten)
        openmvg_keypoint_file_name = path.splitext(path.basename(openmvg_keypoint_file_name))[0] + '.feat'
        openmvg_keypoint_file_path = path.join(openmvg_regions_dir_path, openmvg_keypoint_file_name)
        keypoints_data = image_keypoints_from_file(kapture_keypoint_file_path,
                                                   kapture_keypoints.dtype,
                                                   kapture_keypoints.dsize)
        keypoints_data = keypoints_data[:, 0:4]
        np.savetxt(openmvg_keypoint_file_path, keypoints_data, fmt='%10.5f')

    # copy descriptors files
    """
    from openMVG regions_factory.hpp
    using SIFT_Regions = Scalar_Regions<SIOPointFeature, unsigned char, 128>;
    using AKAZE_Float_Regions = Scalar_Regions<SIOPointFeature, float, 64>;
    using AKAZE_Liop_Regions = Scalar_Regions<SIOPointFeature, unsigned char, 144>;
    using AKAZE_Binary_Regions = Binary_Regions<SIOPointFeature, 64>;
    """
    descriptors = descriptors_to_filepaths(kapture_descriptors, kapture_path)
    for kapture_image_name, kapture_descriptors_file_path in tqdm(descriptors.items(), disable=hide_progress_bars):
        openmvg_descriptors_file_name = get_openmvg_image_path(kapture_image_name, image_path_flatten)
        openmvg_descriptors_file_name = path.splitext(path.basename(openmvg_descriptors_file_name))[0] + '.desc'
        openmvg_descriptors_file_path = path.join(openmvg_regions_dir_path, openmvg_descriptors_file_name)
        kapture_descriptors_data = image_descriptors_from_file(kapture_descriptors_file_path,
                                                               kapture_descriptors.dtype,
                                                               kapture_descriptors.dsize)
        # assign a byte array of [size_t[1] + uint8[nb features x 128]
        size_t_len = 64 // 8
        openmvg_descriptors_data = np.empty(dtype=np.uint8, shape=(kapture_descriptors_data.size + size_t_len,))
        openmvg_descriptors_data[0:size_t_len].view(dtype=np.uint64)[0] = kapture_descriptors_data.shape[0]
        openmvg_descriptors_data[size_t_len:] = kapture_descriptors_data.flatten()
        array_to_file(openmvg_descriptors_file_path, openmvg_descriptors_data)


def export_openmvg_matches(
        kapture_path: str,
        kapture_data: kapture.Kapture,
        openmvg_matches_file_path: str,
        kapture_to_openmvg_view_ids: Dict[str, int]
):
    if kapture_data.matches is None:
        logger.warning('No matches to be exported.')
        return

    if path.splitext(openmvg_matches_file_path)[1] != '.txt':
        logger.warning('Matches are exported as text format, even if file does not ends with .txt.')

    # make sure output directory is ready
    os.makedirs(path.dirname(openmvg_matches_file_path), exist_ok=True)

    hide_progress_bars = logger.getEffectiveLevel() > logging.INFO
    matches = matches_to_filepaths(kapture_data.matches, kapture_path)
    with open(openmvg_matches_file_path, 'w') as fid:
        for image_pair, kapture_matches_filepath in tqdm(matches.items(), disable=hide_progress_bars):
            # idx image1 idx image 2
            # nb pairs
            # pl1 pr1 pl2 pr2 ...
            i, j = [kapture_to_openmvg_view_ids[image_name] for image_name in image_pair]
            fid.write(f'{i} {j}\n')
            matches_indices = image_matches_from_file(kapture_matches_filepath)[:, 0:2].astype(int)
            fid.write(f'{matches_indices.shape[0]}\n')
            for indices_pair in matches_indices:
                fid.write(f'{indices_pair[0]}  {indices_pair[1]}\n')


def export_openmvg(
        kapture_path: str,
        openmvg_sfm_data_file_path: str,
        openmvg_image_root_path: str = None,
        openmvg_regions_dir_path: str = None,
        openmvg_matches_file_path: str = None,
        image_action: TransferAction = TransferAction.skip,
        image_path_flatten: bool = False,
        force: bool = False
) -> None:
    """
    Export the kapture data to an openMVG files.
    If the openmvg_path is a directory, it will create a JSON file (using the default name sfm_data.json)
    in that directory.

    :param kapture_path: full path to input kapture directory
    :param openmvg_sfm_data_file_path: input path to the SfM data file to be written.
    :param openmvg_image_root_path: optional input path to openMVG image directory to be created.
    :param openmvg_regions_dir_path: optional input path to openMVG regions (feat, desc) directory to be created.
    :param openmvg_matches_file_path: optional input path to openMVG matches file to be created.
    :param image_action: an action to apply on the images: relative linking, absolute linking, copy or move. Or top
     directory linking or skip to do nothing. If not "skip" equires openmvg_image_root_path to be defined.
    :param image_path_flatten: flatten image path (eg. to avoid image name collision in openMVG regions).
    :param force: if true, will remove existing openMVG data without prompting the user.
    """

    if any(arg is not None and not isinstance(arg, str)
           for arg in [kapture_path, openmvg_image_root_path, openmvg_regions_dir_path, openmvg_matches_file_path]
           ):
        raise ValueError('expect str (or None) as path argument.')

    # clean before export
    safe_remove_file(openmvg_sfm_data_file_path, force)
    if path.exists(openmvg_sfm_data_file_path):
        raise ValueError(f'{openmvg_sfm_data_file_path} file already exist')

    # load kapture
    logger.info(f'loading kapture {kapture_path}...')
    kapture_data = kapture.io.csv.kapture_from_dir(kapture_path)
    if kapture_data is None or not isinstance(kapture_data, kapture.Kapture):
        raise ValueError(f'unable to load kapture from {kapture_path}')
    kapture_to_openmvg_view_ids = {}

    logger.info(f'exporting sfm data to {openmvg_sfm_data_file_path} ...')
    export_openmvg_sfm_data(
        kapture_data=kapture_data,
        kapture_path=kapture_path,
        openmvg_sfm_data_file_path=openmvg_sfm_data_file_path,
        openmvg_image_root_path=openmvg_image_root_path,
        image_action=image_action,
        image_path_flatten=image_path_flatten,
        force=force,
        kapture_to_openmvg_view_ids=kapture_to_openmvg_view_ids)

    if openmvg_regions_dir_path is not None:
        try:
            logger.info(f'exporting regions to {openmvg_regions_dir_path} ...')
            export_openmvg_regions(
                kapture_path=kapture_path,
                kapture_keypoints=kapture_data.keypoints,
                kapture_descriptors=kapture_data.descriptors,
                openmvg_regions_dir_path=openmvg_regions_dir_path,
                image_path_flatten=image_path_flatten
            )
        except ValueError as e:
            logger.error(e)

    if openmvg_matches_file_path is not None:
        try:
            logger.info(f'exporting matches to {openmvg_matches_file_path} ...')
            export_openmvg_matches(
                kapture_path=kapture_path,
                kapture_data=kapture_data,
                openmvg_matches_file_path=openmvg_matches_file_path,
                kapture_to_openmvg_view_ids=kapture_to_openmvg_view_ids
            )
        except ValueError as e:
            logger.error(e)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Colmap database import as basic kapture objects functions
"""

import logging
import numpy as np
from tqdm import tqdm
from typing import Tuple, Optional

# kapture
import kapture
import kapture.io.features
# local
from .database import COLMAPDatabase, blob_to_array, pair_id_to_image_ids
from .database_extra import exists_table
from .cameras import CAMERA_MODEL_NAMES, get_camera_kapture_id_from_colmap_id

logger = logging.getLogger('colmap')


def get_cameras_from_database(database: COLMAPDatabase) -> kapture.Sensors:
    """
    Creates kapture sensors from the colmap database.

    :param database: colmap database
    :return: kapture sensors
    """
    logger.info('parsing cameras  ...')
    kapture_cameras = kapture.Sensors()

    for camera_id, model_id, width, height, params, prior_focal_length in database.execute(
            'SELECT camera_id, model, width, height, params, prior_focal_length FROM cameras;'):
        if model_id not in CAMERA_MODEL_NAMES:
            logger.warning(f'unable to convert colmap camera model ({model_id}) for camera {camera_id}.')
            # use 0 as default
            model_id = 0

        camera_id = get_camera_kapture_id_from_colmap_id(camera_id)
        model_name = CAMERA_MODEL_NAMES[model_id]

        #  By setting the prior_focal_length flag to 0 or 1,
        #  you can give a hint whether the reconstruction algorithm should trust the focal length value.
        params = blob_to_array(params, np.float64)
        params = [width, height] + params.tolist()

        kapture_camera = kapture.Camera(model_name, params)
        kapture_cameras[camera_id] = kapture_camera
    return kapture_cameras


def get_images_and_trajectories_from_database(database: COLMAPDatabase
                                              ) -> Tuple[kapture.RecordsCamera, kapture.Trajectories]:
    """
    Creates records_camera and trajectories from colmap images table
    In trajectories, timestamps are made up from colmap image id.

    :param database: colmap database
    :return: kapture records_camera and trajectories
    """
    logging.info('parsing images ...')
    kapture_images = kapture.RecordsCamera()
    kapture_trajectories = kapture.Trajectories()
    hide_progressbar = logger.getEffectiveLevel() > logging.INFO
    for image_id, name, camera_id, prior_qw, prior_qx, prior_qy, prior_qz, prior_tx, prior_ty, prior_tz \
            in tqdm(database.execute('SELECT image_id, name, camera_id, '
                                     'prior_qw, prior_qx, prior_qy, prior_qz, '
                                     'prior_tx, prior_ty, prior_tz  FROM images;'
                                     ), disable=hide_progressbar):
        # images
        timestamp = int(image_id)
        camera_id = get_camera_kapture_id_from_colmap_id(camera_id)
        kapture_images[timestamp, camera_id] = name
        # trajectories
        prior_q = [prior_qw, prior_qx, prior_qy, prior_qz]
        prior_t = [prior_tx, prior_ty, prior_tz]
        # do not register the pose part if its invalid.
        is_undefined = all(v is None for v in prior_q + prior_t)
        if is_undefined:
            # just ignore
            continue
        prior_pose = kapture.PoseTransform(prior_q, prior_t)
        kapture_trajectories[timestamp, camera_id] = prior_pose

    if len(kapture_trajectories) == 0:
        # if there is no pose at all, just don't bother.
        kapture_trajectories = None

    return kapture_images, kapture_trajectories


def get_keypoints_from_database(database: COLMAPDatabase,
                                records_camera: kapture.RecordsCamera,
                                kapture_dirpath: str,
                                keypoint_name: str = 'SIFT'
                                ) -> Optional[kapture.Keypoints]:
    """
    Writes keypoints files and return the kapture keypoints from the colmap database.
    Requires records_camera timestamp == colmap_image_id

    :param database: colmap database.
    :param records_camera: input images.
    :param kapture_dirpath: input root path to kapture.
    :param keypoint_name: name of the keypoints detector (by default, in colmap, its SIFT, but can be imported)
    :return: kapture keypoints
    """
    image_filenames = set()
    dtype = np.float32
    dsize = None  # usually 6, will be retrieved on first keypoints of DB

    empty_keypoints = []

    # DB query
    colmap_keypoints_request = (
        (image_id, blob_to_array(data, dtype, (rows, cols)) if (rows > 0 and cols > 0) else None)
        for image_id, rows, cols, data in database.execute("SELECT image_id, rows, cols, data FROM keypoints"))

    hide_progressbar = logger.getEffectiveLevel() > logging.INFO
    for colmap_image_id, image_keypoints in tqdm(colmap_keypoints_request, disable=hide_progressbar):
        # retrieve image path from image_id
        assert len(records_camera[colmap_image_id]) == 1
        timestamp = colmap_image_id
        image_filename = next((v for v in records_camera[timestamp].values()), None)
        assert image_filename
        keypoints_filepath = kapture.io.features.get_keypoints_fullpath(kapture_dirpath, image_filename)

        # handle no keypoints (delay)
        if image_keypoints is None or image_keypoints.shape[0] == 0:
            logger.warning(f'image={image_filename} has 0 keypoints')
            empty_keypoints.append((image_filename, keypoints_filepath))
            continue

        if image_keypoints.dtype != dtype:
            raise ValueError('inconsistent keypoints type.')
        if dsize is None:
            dsize = int(image_keypoints.shape[1])
        elif dsize != image_keypoints.shape[1]:
            raise ValueError('inconsistent keypoints size.')

        # save the actual file
        kapture.io.features.image_keypoints_to_file(keypoints_filepath, image_keypoints)
        # register it into kapture
        image_filenames.add(image_filename)

    # do all the empty keypoints at the end when we know for sure dsize
    for image_filename, keypoints_filepath in empty_keypoints:
        if dsize is None:
            dsize = 6
        image_keypoints = np.zeros((0, dsize), dtype=dtype)
        # save the actual file
        kapture.io.features.image_keypoints_to_file(keypoints_filepath, image_keypoints)
        # register it into kapture
        image_filenames.add(image_filename)

    if image_filenames:
        return kapture.Keypoints(keypoint_name, dtype, dsize, image_filenames)
    else:
        return None


def get_descriptors_from_database(database: COLMAPDatabase,
                                  images: kapture.RecordsCamera,
                                  kapture_dirpath: str,
                                  descriptor_name: str = 'SIFT'
                                  ) -> Optional[kapture.Descriptors]:
    """
    Writes descriptors files and return the list in kapture format from the colmap database.

    :param database: colmap database.
    :param images: list of images (as RecordsCamera).
    :param kapture_dirpath: input root path to kapture.
    :param descriptor_name: name of the keypoints descriptor (by default, in colmap, its SIFT, but can be imported)
    :return: kapture descriptors
    """
    image_filenames = set()
    dtype = np.uint8  # values in the range 0…255
    # see https://colmap.github.io/tutorial.html#feature-detection-and-extraction
    dsize = None  # usually uint8, 128, will be retrieved on first descriptor of DB

    empty_descriptors = []

    colmap_descriptors = (
        (image_id, blob_to_array(data, dtype, (rows, cols)) if (rows > 0 and cols > 0) else None)
        for image_id, rows, cols, data in database.execute("SELECT image_id, rows, cols, data FROM descriptors"))
    hide_progressbar = logger.getEffectiveLevel() > logging.INFO
    for image_id, image_descriptors in tqdm(colmap_descriptors, disable=hide_progressbar):
        # retrieve image path from image_id (actually the timestamp)
        image_filename = next((v for v in images[image_id].values()), None)
        assert image_filename
        descriptors_filepath = kapture.io.features.get_descriptors_fullpath(kapture_dirpath, image_filename)

        if image_descriptors is None or image_descriptors.shape[0] == 0:
            logger.warning(f'image={image_id}:{image_filename} has 0 descriptors.')
            empty_descriptors.append((image_filename, descriptors_filepath))
            continue

        if image_descriptors.dtype != dtype:
            raise ValueError('inconsistent descriptors type.')
        if dsize is None:
            dsize = int(image_descriptors.shape[1])
        elif dsize != image_descriptors.shape[1]:
            raise ValueError('inconsistent descriptors size.')

        # save the actual file
        kapture.io.features.image_descriptors_to_file(descriptors_filepath, image_descriptors)
        # register it into to kapture
        image_filenames.add(image_filename)

    # do all the empty descriptors at the end when we know for sure dsize
    for image_filename, descriptors_filepath in empty_descriptors:
        if dsize is None:
            dsize = 128
        image_descriptors = np.zeros((0, dsize), dtype=dtype)
        # save the actual file
        kapture.io.features.image_descriptors_to_file(descriptors_filepath, image_descriptors)
        # register it into kapture
        image_filenames.add(image_filename)

    if image_filenames:
        return kapture.Descriptors(descriptor_name, dtype, dsize, image_filenames)
    else:
        return None


def get_matches_from_database(database: COLMAPDatabase,
                              images: kapture.RecordsCamera,
                              kapture_dirpath: str,
                              no_geometric_filtering: bool) -> kapture.Matches:
    """
    Writes Matches files and return the list in kapture format from the colmap database.

    :param database: input colmap database.
    :param images: input list of images (as RecordsCamera).
    :param kapture_dirpath: input root path to kapture.
    :param no_geometric_filtering: only retrieve matches with geometric consistency.
    :return: kapture matches
    """
    kapture_matches = kapture.Matches()
    # check there is geometric matches available
    matches_table_name = 'matches'
    if not no_geometric_filtering:
        if not exists_table('two_view_geometries', database):
            logger.warning("No table: two_view_geometries: skipping geometric filtering")
        else:
            request = database.execute(
                'SELECT COUNT (*) FROM two_view_geometries')
            nb_verified_matches = next(request)[0]
            if nb_verified_matches > 0:
                request = database.execute('SELECT COUNT (*) FROM matches')
                nb_total_matches = next(request)[0]
                logger.info('keeps {}% of verified matches ({}/{}) ...'.format(
                    nb_verified_matches / nb_total_matches * 100, nb_verified_matches, nb_total_matches))
                matches_table_name = 'two_view_geometries'

    colmap_matches = [(pair_id_to_image_ids(pair_id), blob_to_array(data, np.uint32, (rows, cols)))
                      if rows > 0 else (pair_id_to_image_ids(pair_id), np.empty((rows, cols), dtype=np.uint32))
                      for pair_id, rows, cols, data
                      in database.execute('SELECT pair_id, rows, cols, data FROM {}'.format(matches_table_name))]

    logger.debug('matches: {}'.format(len(colmap_matches)))
    hide_progressbar = logger.getEffectiveLevel() > logging.INFO
    for (image_id1, image_id2), image_matches in tqdm(colmap_matches, disable=hide_progressbar):
        if image_id1 not in images or image_id2 not in images:
            logger.critical('inconsistent image ID {} or {}'.format(image_id1, image_id2))
            continue
        filename1 = next((v for v in images[image_id1].values()), None)
        filename2 = next((v for v in images[image_id2].values()), None)
        assert filename1 and filename2
        if (filename1, filename2) != kapture.Matches.lexical_order(filename1, filename2):
            # have to swap matches (keypoint image1, keypoint image2) become (keypoint image2, keypoint image1)
            image_matches = image_matches[:, ::-1]
            filename1, filename2 = kapture.Matches.lexical_order(filename1, filename2)

        # actually write the file
        # convert colmap image matches into kapture (cast to float and add a score column)
        image_matches = image_matches.astype(np.float)
        image_matches = np.hstack([image_matches, np.zeros((image_matches.shape[0], 1))])
        image_matches_filepath = kapture.io.features.get_matches_fullpath((filename1, filename2), kapture_dirpath)
        kapture.io.features.image_matches_to_file(image_matches_filepath, image_matches)
        # register the matching in kapture
        kapture_matches.add(filename1, filename2)

    return kapture_matches



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from typing import Dict, List, Tuple


class Observations(Dict[int, List[Tuple[str, int]]]):
    """
    Observations. This can be used like this:

    - observations[point3d_idx]= list( observation )
    - observation = (image_path, keypoint_idx)
    """

    def add(self, point3d_idx: int, image_filename: str, keypoint_idx: int):
        """
        Adds a 2-D observation (image, keypoint) of a 3D point.

        :param point3d_idx: index of the 3D point to add an observation of.
        :param image_filename: name of the image where the 3D points is observed
        :param keypoint_idx: index of the keypoints in the image that correspond to the 3D point.
        :return:
        """
        # enforce type checking
        if not isinstance(point3d_idx, int):
            raise TypeError('invalid type for point3d_idx')
        if not isinstance(image_filename, str):
            raise TypeError('invalid type for image_filename')
        if not isinstance(keypoint_idx, int):
            raise TypeError('invalid type for keypoint_idx')
        self.setdefault(point3d_idx, []).append((image_filename, keypoint_idx))

    def __repr__(self):
        representation = ''
        for point3d_idx in sorted(self.keys()):
            representation += f'[{point3d_idx:05}]: '
            for image_path, keypoint_idx in self.get(point3d_idx):
                representation += f'\t({image_path}, {keypoint_idx})'
            representation += '\n'
        return representation



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Functions to export the reconstruction part of colmap (only text format is supported).
"""

import logging
import os.path as path
from typing import Dict, Tuple, List

# kapture
import kapture
import kapture.io.features
# local
from .cameras import CAMERA_MODEL_NAMES, get_colmap_camera

logger = logging.getLogger('colmap')


def export_to_colmap_cameras_txt(colmap_cameras_filepath: str,
                                 sensors: kapture.Sensors,
                                 colmap_camera_ids: Dict[str, int]) -> None:
    """
    Exports to colmap reconstruction file "cameras.txt".

    :param colmap_cameras_filepath: path to colmap file "cameras.txt" to be writen.
    :param sensors: sensors to be exported
    :param colmap_camera_ids: gives the correspondences between kapture camera id and colmap camera id
    """
    assert path.basename(colmap_cameras_filepath) == 'cameras.txt'
    assert isinstance(colmap_camera_ids, dict)
    assert isinstance(sensors, kapture.Sensors)
    cameras = {cam_id: cam
               for cam_id, cam in sensors.items()
               if isinstance(cam, kapture.Camera)}

    # cameras.txt
    cameras_colmap_header = '# Sensor list with one line of data per camera:\n' \
                            '#   CAMERA_ID, MODEL, WIDTH, HEIGHT, PARAMS[]\n' \
                            '# NB cameras : {}\n'.format(len(cameras))

    converted_cameras = [(cam_id, get_colmap_camera(cam)) for cam_id, cam in cameras.items()]
    with open(colmap_cameras_filepath, 'w') as fid:
        fid.write(cameras_colmap_header)
        lines = (' '.join([str(colmap_camera_ids[cam_id]), CAMERA_MODEL_NAMES[col_cam_id]] +
                          [str(int(width)), str(int(height))] +
                          [str(int(v)) if v.is_integer() else str(v) for v in params])
                 for cam_id, (col_cam_id, width, height, params, prior_focal_length) in converted_cameras
                 if cam_id in colmap_camera_ids)
        fid.write('\n'.join(lines))


def export_to_colmap_images_txt(colmap_images_filepath: str,
                                images: kapture.RecordsCamera,
                                trajectories: kapture.Trajectories,
                                colmap_camera_ids: Dict[str, int],
                                colmap_image_ids: Dict[str, int],
                                image_to_keypoints: Dict[str, List[Tuple[float, float, int]]]) -> None:
    """
    Exports kapture to colmap reconstruction file "images.txt".

    :param colmap_images_filepath: path to colmap file "images.txt" to be writen.
    :param images: images list to export
    :param trajectories: poses to export
    :param colmap_camera_ids: correspondences between kapture camera id and colmap camera id
    :param colmap_image_ids: correspondences between kapture image id (image path) and colmap image id
    :param image_to_keypoints: input image_filename -> [(x, y, point_id), (x, y, point_id), ...]
    """
    assert path.basename(colmap_images_filepath) == 'images.txt'
    assert isinstance(images, kapture.RecordsCamera)
    assert isinstance(trajectories, kapture.Trajectories)
    assert isinstance(colmap_camera_ids, dict)
    assert isinstance(colmap_image_ids, dict)
    assert isinstance(image_to_keypoints, dict)

    images_flattened = list(kapture.flatten(images))
    images_colmap_header = '# Image list with two lines of data per image:\n' \
                           '#   IMAGE_ID, QW, QX, QY, QZ, TX, TY, TZ, CAMERA_ID, NAME\n' \
                           '#   POINTS2D[] as (X, Y, POINT3D_ID)\n' \
                           '# NB IMAGES : {}\n'.format(len(images_flattened))

    with open(colmap_images_filepath, 'w') as fid:
        fid.write(images_colmap_header)
        for timestamp, sensor_id, sensing_filepath in images_flattened:
            colmap_cam_id = colmap_camera_ids[sensor_id]
            colmap_image_id = colmap_image_ids[sensing_filepath]
            # retrieve image pose from trajectories

            if timestamp not in trajectories:
                logging.debug('timestamp:{} not in trajectories'.format(timestamp))
                continue

            if sensor_id not in trajectories[timestamp]:
                logger.debug('camera {} not found in trajectories for timestamp {}.'.format(
                    sensor_id, timestamp))
                continue

            pose_tr = trajectories[timestamp].get(sensor_id)
            line = [colmap_image_id] + pose_tr.r_raw + pose_tr.t_raw + [colmap_cam_id, sensing_filepath]
            fid.write(' '.join('{}'.format(i) for i in line) + '\n')
            if sensing_filepath in image_to_keypoints:
                # POINTS2D[] as (X, Y, POINT3D_ID)
                p2d = [(str(x), str(y), str(point_id))
                       for x, y, point_id in image_to_keypoints[sensing_filepath]]
                fid.write(' '.join('{} {} {}'.format(i, j, k) for i, j, k in p2d))
            fid.write('\n')


def export_to_colmap_matches_txt(colmap_matches_filepath: str, matches: kapture.Matches) -> None:
    """
    Exports to colmap reconstruction file "matches.txt". This files can be used to redo the geometric filtering.

    :param colmap_matches_filepath: input path to colmap matches file to be writen.
    :param matches: input kapture.Matches
    """
    assert isinstance(matches, kapture.Matches)
    with open(colmap_matches_filepath, 'w') as fid:
        # matches[(image_path1, image_path2)] = image_matches
        for image_path1, image_path2 in matches:
            fid.write('{} {}\n'.format(image_path1, image_path2))


def export_to_colmap_points3d_txt(colmap_points3d_filepath: str,
                                  colmap_image_ids: Dict[str, int],
                                  points3d: kapture.Points3d = None,
                                  observations: kapture.Observations = None) -> None:
    """
    Exports to colmap points3d text file.

    :param colmap_points3d_filepath: path to colmap points3d file to be writen.
    :param colmap_image_ids: correspondences between kapture image id (image path) and colmap image id
    :param points3d: kapture points3d to export
    :param observations: kapture observations to export
    """
    assert isinstance(points3d, kapture.Points3d) or points3d is None
    assert isinstance(observations, kapture.Observations) or observations is None
    assert isinstance(colmap_image_ids, dict)
    points3d_colmap_header = '# 3D point list with one line of data per point:\n' \
                             '#   POINT3D_ID, X, Y, Z, R, G, B, ERROR, TRACK[] as (IMAGE_ID, POINT2D_IDX)\n'
    with open(colmap_points3d_filepath, 'w') as fid:
        fid.write(points3d_colmap_header)
        if points3d:
            for i in range(points3d.shape[0]):
                point3d = points3d[i]
                line = '{} {} {} {} {} {} {} 0'.format(i,
                                                       point3d[0], point3d[1], point3d[2],
                                                       int(point3d[3]), int(point3d[4]), int(point3d[5]))
                if observations is not None and i in observations and len(observations[i]) > 0:
                    line += ' '
                    pairs = [(str(colmap_image_ids[name]), str(keypoint_id)) for name, keypoint_id in observations[i]]
                    line += ' '.join([str(s) for s in list(sum(pairs, ()))])
                line += '\n'
                fid.write(line)


def export_to_colmap_txt(colmap_reconstruction_dirpath: str,
                         kapture_data: kapture.Kapture,
                         kapture_dirpath: str,
                         colmap_camera_ids: Dict[str, int],
                         colmap_image_ids: Dict[str, int]) -> None:
    """
    Exports to colmap reconstruction text files.

    :param colmap_reconstruction_dirpath: path to directory where colmap reconstruction files will be stored.
    :param kapture_data: input kapture data
    :param kapture_dirpath: path to output directory, where colmap files will be stored
    :param colmap_camera_ids: gives the correspondences between kapture camera id and colmap camera id
    :param colmap_image_ids: gives the correspondences between kapture image id (image path) and colmap image id
    """
    assert isinstance(kapture_data, kapture.Kapture)
    assert isinstance(colmap_camera_ids, dict)
    assert isinstance(colmap_image_ids, dict)
    assert kapture_data.records_camera is not None
    assert kapture_data.sensors is not None

    if kapture_data.rigs:
        # check the rigs are not used in trajectories
        used_sensors = set(sensor_id
                           for shot in kapture_data.trajectories.values()
                           for sensor_id in shot.keys())
        rigs_ids = set(rig_id for rig_id in kapture_data.rigs.keys())
        if any(used_sensor in rigs_ids for used_sensor in used_sensors):
            raise ValueError(
                'colmap format does not handle rigs notation. '
                'Remove rig from trajectories beforehand (see rigs_remove_inplace)')

    # cameras.txt
    logger.info('creating colmap cameras.txt')
    export_to_colmap_cameras_txt(path.join(colmap_reconstruction_dirpath, 'cameras.txt'),
                                 kapture_data.sensors, colmap_camera_ids)

    # images.txt
    image_to_keypoints = {}
    if kapture_data.keypoints and kapture_data.points3d and kapture_data.observations:
        observations_reversed = {(image_filename, keypoint_idx): point3d_idx
                                 for point3d_idx, (image_filename, keypoint_idx) in
                                 kapture.flatten(kapture_data.observations)}

        # prepare images.txt even lines
        # POINTS2D[] as (X, Y, POINT3D_ID)
        keypoints_filepaths = kapture.io.features.keypoints_to_filepaths(kapture_data.keypoints, kapture_dirpath)
        for image_filename, image_keypoints_filepath in keypoints_filepaths.items():
            image_keypoints = kapture.io.features.image_keypoints_from_file(image_keypoints_filepath,
                                                                            kapture_data.keypoints.dtype,
                                                                            kapture_data.keypoints.dsize)
            image_to_keypoints[image_filename] = []
            for i in range(image_keypoints.shape[0]):
                point3d_idx = observations_reversed[(image_filename, i)] if (image_filename,
                                                                             i) in observations_reversed else -1
                image_to_keypoints[image_filename].append((image_keypoints[i, 0], image_keypoints[i, 1], point3d_idx))

    if kapture_data.records_camera is None or kapture_data.trajectories is None:
        logger.info('skipping colmap images.txt (missing images or trajectories).')
    else:
        logger.info('creating colmap images.txt')
        export_to_colmap_images_txt(path.join(colmap_reconstruction_dirpath, 'images.txt'),
                                    kapture_data.records_camera, kapture_data.trajectories,
                                    colmap_camera_ids,
                                    colmap_image_ids,
                                    image_to_keypoints)

    # image_matches.txt: to be imported as custom match (allow geometric verification)
    if kapture_data.matches:
        logger.info('creating image_matches.txt')
        export_to_colmap_matches_txt(path.join(colmap_reconstruction_dirpath, 'image_matches.txt'),
                                     kapture_data.matches)

    # points3D.txt
    logging.info('creating colmap points3D.txt')
    export_to_colmap_points3d_txt(path.join(colmap_reconstruction_dirpath, 'points3D.txt'), colmap_image_ids,
                                  kapture_data.points3d, kapture_data.observations)
